from imports import *
from plotly_graphs import *
from protodash import *
from insights import *
from plotly_css import *
import pandasql as psql
import string
import random
import os
from apps import global_explanation, local_explanation, distribution, feature_interaction
from app import app
from what_if import *
from calculate_shap import *

class dashboard():
    def __init__(self):
        super(dashboard, self).__init__()
        self.param = None
        self.query_dict = dict()
        self.filtered_dataframe = dict()
        self.feature_importance = dict()
        self.pdp = dict()
        self.summary_plot = dict()
        self.feature_impact = dict()
        self.multi_level_eda = dict()


    def create_dir(self, dir_name):
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)

    def random_string_generator(self):
        random_str = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(10))
        return random_str

    def caching_data_exist_in_dict(self, sql_query, graph_type):
        if graph_type == 'feature_importance':
            if sql_query in self.feature_importance:
                return True, self.feature_importance[sql_query]
            return False, None
        elif graph_type == 'pdp':
            if sql_query in self.pdp:
                return True, self.pdp[sql_query]
            return False, None
        elif graph_type == 'summary_plot':
            if sql_query in self.summary_plot:
                return True, self.summary_plot[sql_query]
            return False, None
        elif graph_type == 'feature_impact':
            if sql_query in self.feature_impact:
                return True, self.feature_impact[sql_query]
            return False, None
        elif graph_type == 'multi_level_eda':
            if sql_query in self.multi_level_eda:
                return True, self.multi_level_eda[sql_query]
            return False, None
        elif graph_type == 'filtered_df':
            if sql_query in self.query_dict:
                return True, self.query_dict[sql_query]
            return False, None

    def caching_exists_in_file(self, sql_query, graph_type):
        self.create_dir("data_storage")
        self.create_dir("data_storage/{}".format(graph_type))

        try:
            dictionary_csv = pd.read_csv("./data_storage/filtered_df/dictionary_bkp.csv")
            random_id = dictionary_csv[(dictionary_csv['sql'] == sql_query) & (dictionary_csv['type'] == graph_type)
                                       & (dictionary_csv['instance_id'] == self.instance_id)]

            random_id.drop_duplicates(['sql'], keep='last', inplace=True)
            if not random_id.empty:
                result = random_id['random_id'].iloc[0]
                dff = pd.read_csv("./data_storage/{}/{}.csv".format(graph_type, result))
                #print("{}  {}  {}".format(graph_type, "file exists", result))
                return True, True, dff
            return False, True, None

        except Exception as e:
            return False, False, None

    def creating_filtered_backup_file(self, sql_query, random_str, graph_type):
        dict_bkp = pd.DataFrame(data={"sql": [sql_query], "random_id": [random_str], "type": [graph_type],
                                      "instance_id": [self.instance_id]})
        return dict_bkp

    def store_data_in_csv(self, df, graph_type, random_str):
        df.to_csv("./data_storage/{}/{}.csv".format(graph_type, random_str), index=False)

    def store_data_in_dict(self, df, sql_query, graph_type):
        if graph_type == 'feature_importance':
            self.feature_importance[sql_query] = df
        elif graph_type == 'pdp':
            self.pdp[sql_query] = df
        elif graph_type == 'summary_plot':
            self.summary_plot[sql_query] = df
        elif graph_type == 'feature_impact':
            self.feature_impact[sql_query] = df
        elif graph_type == 'multi_level_eda':
            self.multi_level_eda[sql_query] = df
        elif graph_type == 'filtered_df':
            self.query_dict[sql_query] = df

    def caching_data_manager(self, df, sql_query, graph_type, calculated_funct=None, details_dict=None):
        status_file, file_exist, dff = self.caching_exists_in_file(sql_query, graph_type)
        if status_file:
            #print("{}/{}".format(graph_type, "exist in file"))

            return dff
        else:
            #print("{}/{}".format(graph_type, "don't exists"))
            random_str = self.random_string_generator()
            dict_bkp = self.creating_filtered_backup_file(sql_query, random_str, graph_type)
            dff = psql.sqldf(sql_query, locals())
            self.create_dir("data_storage/filtered_df")
            dict_bkp.to_csv("./data_storage/{}/dictionary_bkp.csv".format("filtered_df"), mode='a',
                            header=file_exist is False,
                            index=False)
            dff.to_csv("./data_storage/filtered_df/{}.csv".format(random_str), mode='w',
                       header=file_exist is False,
                       index=False)

            results = dff
            if graph_type != 'filtered_df':
                results = calculated_funct(dff)
            self.store_data_in_csv(results, graph_type, random_str)
            return results

    def find(self, df, mode, param):
        self.available_columns = available_columns = list(df.columns)
        original_variables = [col for col in df.columns if '_impact' in col]
        self.impact_variables = [col for col in original_variables if not '_rescaled' in col]
        self.y_variable = param["y_variable"]
        self.y_variable_predict = param["y_variable_predict"]
        self.param = param
        self.instance_id = self.random_string_generator()
        self.create_dir("data_storage")
        self.create_dir("data_storage/user")
        self.user_id = None
        self.df=df

        self.callback_input = [Input(f + '_slider', 'value') for f in self.param["columns"]]
        self.callback_input.append(Input('submit-button-state', 'n_clicks'))

        # self.callback_input_prototype = [Input(f + '-slider', 'value') for f in self.param["columns"]]
        # self.callback_input_prototype.append(Input('btn-nclicks-1', 'n_clicks'))

        self.prototype_array= []
        for f in self.param["columns"]:
            self.prototype_array.append([f + '_slider', 'value'])
        self.prototype_array.append(['btn-nclicks-1', 'n_clicks'])
        try:
            user_id = pd.read_csv("data_storage/user/user_id.csv")
            user_id.drop_duplicates(['id'], keep='first', inplace=True)
            user_id = user_id['id'].iloc[0]
            self.user_id = user_id
        except Exception as e:
            #print("inside user track" )
            user_id_val = self.random_string_generator()
            user_id_csv = pd.DataFrame(data={"id": [user_id_val]})
            user_id_csv.to_csv("data_storage/user/user_id.csv", index=False)
            self.user_id= user_id_val  
        self.insights = insights(self.param)
        d = self.dash(df, mode)
        return True

    def dash(self, df, mode):
        y_variable = self.y_variable
        g = plotly_graphs()
        y_variables = [col for col in df.columns if '_impact' in col]
        original_variables = [col for col in df.columns if not '_impact' in col]
        original_variables = [col for col in original_variables if not '_rescaled' in col]
        array = []
        PAGE_SIZE = 10
        row_number = 1
        # y_variable = "predict"
        columns = ['index', '0', '1', '2', '3', '4']
        dat = []

        available_columns = list(df.columns)

        external_stylesheets = ['https://codepen.io/rab657/pen/LYpKraq.css',
                                {
                                    'href': 'https://stackpath.bootstrapcdn.com/bootstrap/4.5.0/css/bootstrap.min.css',
                                    'rel': 'stylesheet'
                                }
                                ]

        # app = JupyterDash(__name__, external_stylesheets=external_stylesheets)

        app.title = "explainX.ai - Main Dashboard"

        PLOTLY_LOGO = "https://images.plot.ly/logo/new-branding/plotly-logomark.png"

        menu = dbc.Row(
            [
                dbc.Col(dbc.NavItem(dbc.NavLink("Global Explanation", href="/apps/global_explanation")), style={'width':"200px",'fontSize':'12px'}),
                dbc.Col(dbc.NavItem(dbc.NavLink("Local Explanation", href="/apps/local_explanation")), style={'width':"200px",'fontSize':'12px'}),
                dbc.Col(dbc.NavItem(dbc.NavLink("Feature Interaction", href="/apps/feature_interaction")), style={'width':"200px",'fontSize':'12px'}),
                dbc.Col(dbc.NavItem(dbc.NavLink("Distributions", href="/apps/distribution")), style={'width':"200px",'fontSize':'12px'}),
                
            ],
            no_gutters=True,
            className="ml-auto flex-nowrap mt-3 mt-md-0",
            align="center"
        )

        navbar = dbc.Navbar(
            [
                html.A(
                    # Use row and col to control vertical alignment of logo / brand
                    dbc.Row(
                        [
                            dbc.Col(html.Img(src=PLOTLY_LOGO, height="30px")),
                            dbc.Col(dbc.NavbarBrand("explainX.ai", className="ml-2", style={'fontSize':'12px','color':'black'})),
                        ],
                        align="center",
                        no_gutters=True,
                    ),
                    href="https://www.explainx.ai",
                ),
                
                dbc.NavbarToggler(id="navbar-toggler"),
                dbc.Collapse(menu, id="navbar-collapse", navbar=True),
            ],
            color="light",
            dark=True,
        )


        # add callback for toggling the collapse on small screens
        @app.callback(
            Output("navbar-collapse", "is_open"),
            [Input("navbar-toggler", "n_clicks")],
            [State("navbar-collapse", "is_open")],
        )
        def toggle_navbar_collapse(n, is_open):
            if n:
                return not is_open
            return is_open

        @app.callback(Output('page-content', 'children'),
              [Input('url', 'pathname')])
        def display_page(pathname):
            if pathname == '/apps/global_explanation':
                return global_explanation.global_explanation(original_variables)
            elif pathname == '/apps/feature_interaction':
                return feature_interaction.layout_interaction(original_variables, y_variables)
            elif pathname == '/apps/distribution' :
                return distribution.layout_distribution(original_variables)
            elif pathname == '/apps/local_explanation':
                return local_explanation.layout_local(original_variables,columns,df.columns)
            else:
                return welcome_message


        welcome_message= html.Div(
            [
                html.H1("Welcome to ExplainX"),
                html.H3("Click on one of the tabs above to start explaining.")
            ]


        )
        app.layout = html.Div([
            navbar,
            html.Div([
                dbc.Card(
                    [
                        dbc.CardHeader(
                            html.H2(
                                dbc.Button(
                                    "Analyze using SQL",
                                    id="collapse-button-2",
                                    color="link",
                                    style={'fontSize': '14px'})
                            )
                        ),
                        dbc.Collapse(html.Div([
                            html.Div(dcc.Input(
                                id='input-on-submit',
                                type='text',
                                # placeholder="SELECT * FROM df",
                                value="SELECT * FROM df",
                                style={'height': '200px', 'width': '700px', 'fontSize': '15px'})),

                            html.Button('Execute Query', id='submit-val', n_clicks=1),
                            html.Div(id='sql-query-button',
                                     children='Enter a value and press submit',
                                     style={'display': 'none'}
                                     )

                        ], style={'marginTop': 0}), id="collapse-2"),
                    ]
                ),

            ], style=style4),

            html.Div([
                dbc.Card(
                    [
                        dbc.CardHeader(
                            html.H2(
                                dbc.Button(
                                    "View Your Data",
                                    id="collapse-button",
                                    color="link",
                                    style={'fontSize': '14px'})
                            )
                        ),
                        dbc.Collapse(html.Div([
                            html.H4('',
                                    style=style1),
                            html.Div([
                                dash_table.DataTable(
                                    id='datatable-interactivity',
                                    columns=[
                                        {"name": i, "id": i, "deletable": False, "selectable": True} for i in df.columns
                                    ],
                                    # data=df.to_dict('records'),
                                    editable=True,

                                    sort_mode="multi",

                                    # selected_columns=[],
                                    # selected_rows=[],
                                    # page_action="native",
                                    page_current=0,
                                    page_size=PAGE_SIZE,
                                    row_selectable='multi',
                                    page_action='custom',
                                    style_table=style2,
                                    style_header=style3,

                                    style_cell={
                                        "font-family": "Helvetica, Arial, sans-serif",
                                        "fontSize": "11px",
                                        'width': '{}%'.format(len(df.columns)),
                                        'textOverflow': 'ellipsis',
                                        'overflow': 'hidden',
                                        'textAlign': 'left',
                                        'minWidth': '180px', 'width': '180px', 'maxWidth': '180px',
                                    },
                                    css=[
                                        {
                                            'selector': '.dash-spreadsheet td div',

                                            'rule': '''
                                    line-height: 15px;
                                    max-height: 20px; min-height: 20px; height: 20px;
                                    display: block;
                                    overflow-y: hidden;

                                '''
                                        }])
                            ])

                        ], style={'marginTop': 0}), id="collapse"),
                    ]
                ),

            ],
                style=style4),
            #Pages Data
            html.Div([
                dcc.Location(id='url', refresh=False),
                html.Div(id='page-content')
            ])
            # end of collapsable div
           ], style=style40)

         #Navigation
        
       
        

       #Data Interactivity
        @app.callback(
            Output('datatable-interactivity', 'data'),
            [Input('datatable-interactivity', "page_current"),
             Input('datatable-interactivity', "page_size")])
        def update_table(page_current, page_size):
            return df.iloc[
                   page_current * page_size:(page_current + 1) * page_size
                   ].to_dict('records')
        #Collapse-Toggle
        @app.callback(
            Output("collapse", "is_open"),
            [Input("collapse-button", "n_clicks")],
            [State("collapse", "is_open")],
        )
        def toggle_collapse(n, is_open):
            if n:
                return not is_open
            return is_open

        #Collapse-Toggle 2
        @app.callback(
            Output("collapse-2", "is_open"),
            [Input("collapse-button-2", "n_clicks")],
            [State("collapse-2", "is_open")],
        )
        def toggle_collapse(n, is_open):
            if n:
                return not is_open
            return is_open

        #SQL - Data Input Callback
        @app.callback(
            dash.dependencies.Output('sql-query-button', 'children'),
            [dash.dependencies.Input('submit-val', 'n_clicks')],
            [dash.dependencies.State('input-on-submit', 'value')])
        def update_output(n_clicks, value):
            sql_query = f'{value}'
            return sql_query
            # value1 = 'SELECT * FROM df'
            # if n_clicks:
               
            # else:
            #     return value1
        
        #What-If Form CallBack
        @app.callback(
        Output('place_form_here', 'children'),
        [
            # Input('submit-button-state', 'n_clicks'),
        Input('row_number','value')])
        def create_what_if_form(row_number):
            x = what_if()
            i = 0
            if type(row_number) == type(1):
                i = row_number
            array = df[i:i + 1]
            array1 = array
            impact_variables = [col for col in array if '_impact' in col]
            for col in impact_variables:
                array1.drop([col], axis=1, inplace=True)
            # features = [col for col in array if not '_impact' in col]
            features= list(self.param["columns"])
            features.append("y_prediction")
            features.append("y_actual")

            form = x.what_if_form(array1, features)
            return form    

        """
        change this. Take input from what-if form.
        """
        # Local Feature Impact Graph
        @app.callback(
            [Output('local_feature_impact', "figure"),
                Output('local_message_1', "children"),
                Output('local_message_2', "children"),
                Output('local_message_3', "children")],
            self.callback_input)
        def update_impact_graph(*values):
            changed_id = [p['prop_id'] for p in dash.callback_context.triggered][0]
            # if 'submit-button-state' in changed_id:


            # use values to construct a dataframe
            # find prediction and shap values.

            df= pd.DataFrame([values[:-1]])
            df.columns= self.param["columns"]
            # print(df)

            #find prediction and impact values first here please.
            array = self.calculate_prediction_shap(df)


            figure, dat = g.feature_impact_old(array)
            # print(dat)
            message = self.insights.insight_2_local_feature_impact(dat)
            return figure, message[0], message[1], message[2]
            # else:
            #     df2 = self.df[0:1]


            #     figure, dat = g.feature_impact_old(df2)
            #     # print(dat)
            #     message = self.insights.insight_2_local_feature_impact(dat)
            #     return figure, message[0], message[1], message[2]

        # Prototypical Analysis
        """
        Change this. Take input from what-if from
        """
        @app.callback(
            Output(component_id='prototype_data', component_property='data'),
            [Input(f[0], f[1])for f in self.prototype_array])
        def update_table(*values):
            changed_id = [p['prop_id'] for p in dash.callback_context.triggered][0]
            if 'btn-nclicks-1' in changed_id:
                #get x variables and prediction column from the data

                df_row = pd.DataFrame([values[:-1]])
                df_row.columns = self.param["columns"]


                # find prediction and impact values first here please.
                df_row = self.calculate_prediction(df_row)
                df_selected= df[list(self.param["columns"])+ [self.param["y_variable_predict"]]]

                row_number = len(df_selected)
                df_selected.loc[row_number] = df_row.values[0]

                p = protodash()
                p.preprocess_data(df_selected, self.param["y_variable_predict"])




                dfs, sample_row = p.find_prototypes(row_number)
                dat = dfs.T.reset_index()
                print("sample row columns")
                sample_row = sample_row.to_frame()
                sample_row.rename(columns={sample_row.columns[0]: "orig"}, inplace=True)
                sample_row.reset_index(inplace=True)
                print(sample_row.columns)
                dat = pd.merge(dat, sample_row, on=['index'], how='left')
                dat['orig'] = dat['orig'].astype(float)
                for i in list(dat.columns):
                    dat[str(i) + '_color'] = np.nan
                    if i != 'index':
                        dat[i] = dat[i].astype(float)
                        dat[str(i) + '_color'] = dat[str(i) + '_color'].astype(float)
                    dat[str(i) + '_color'] = np.where(dat['orig'] == dat[i], 1, 0)
                dat.drop(["index_color","orig_color"], axis=1, inplace=True)
                dat = dat.to_dict('records')
                return dat
            else:
                return []
        
        # Global Feature Importance
        @app.callback(
            [Output('global_feature_importance', "figure"),
                Output('global_message_1', "children"),
                Output('global_message_2', "children"),
                Output('global_message_3', "children"),
                Output('global_message_4', "children"),
                Output('global_message_5', "children")],
            [Input('sql-query-button', 'children'),
            Input('xaxis-column-test-2', 'value')])
        def update_graphs(sql_query, value ):
            g = plotly_graphs()
            graph_type = "feature_importance"
            dff = self.caching_data_manager(df, sql_query, graph_type, g.feature_importance)
            message = self.insights.insight_1_feature_imp(dff)
            figure = g.feature_importance_graph(dff)
           
            if len(message) == 4:
                return figure, message[0], message[1], message[2], message[3], ""
            return figure, message[0], message[1], message[2], message[3], message[4]

        # Global Feature Impact
        @app.callback(
            [Output('global_feature_impact', "figure"),
                Output('message_1', "children"),
                Output('message_2', "children"),
                Output('message_3', "children")],
            [Input('sql-query-button', 'children'),
            Input('xaxis-column-test-2', 'value')])
        def update_graphs(sql_query,value):
            g = plotly_graphs()
            graph_type = "feature_impact"
            df3 = self.caching_data_manager(df, sql_query, graph_type, g.feature_impact)
            figure = g.feature_impact_graph(df3)
            message = self.insights.insight_2_global_feature_impact(df3)
            return figure, message[0], message[1], message[2]
      
        # Partial Dependence Plot Graph
        @app.callback(
            Output('indicator-graphic', 'figure'),
            [Input('xaxis-column', 'value'),
                Input('yaxis-column', 'value'),
                Input('third-axis', 'value'),
                Input('sql-query-button', 'children')])
        def update_graph(xaxis_column_name, yaxis_column_name, third_axis_name, sql_query):
            g = plotly_graphs()
            graph_type = 'pdp'
            df3 = self.caching_data_manager(df, sql_query, graph_type, g.partial_dependence_plot)
            fig = g.pdp_plot(df3, df3[xaxis_column_name], df3[yaxis_column_name],
                                df3[third_axis_name])
            return fig
           

        # Summary Plot
        @app.callback(
            Output('summary_plot', 'figure'),
            [Input('sql-query-button', 'children'),
            Input('xaxis-column-test', 'value')])
        def update_graph2(sql_query, value):
            g = plotly_graphs()
            graph_type = 'summary_plot'
            df3 = self.caching_data_manager(df, sql_query, graph_type, g.summary_plot)
            print(df3)
            fig = g.summary_plot_graph(df3)
            return fig

        # Distributions
        @app.callback(
            Output('indicator-graphic2', 'figure'),
            [Input('xaxis-column-name', 'value'),
             Input('plot_type', 'value'),
             Input('sql-query-button', 'children'),

             ])
        def update_graph2(xaxis_column_name, plot_type, sql_query):
            graph_type = 'filtered_df'
            df3 = self.caching_data_manager(df, sql_query, graph_type)
            cat_variables = []
            num_variables = []
            for i in original_variables:
                if df[i].dtype == 'object':
                    cat_variables.append(i)
                else:
                    num_variables.append(i)

            if plot_type == "Histogram":
                # group_labels = ['xxaxis_column_name']
                return px.histogram(df3, x=xaxis_column_name, marginal="box")
            else:
                for i in cat_variables:
                    return px.violin(df3, x=xaxis_column_name, box=True, points='all')
                else:
                    return px.violin(df3, y=xaxis_column_name, box=True, points='all', )

        # Multi Level EDA
        @app.callback(
            Output('multi_level_eda', 'figure'),
            [Input('x_axis', 'value'),
             Input('y_axis', 'value'),
             Input('size', 'value'),
             Input('color', 'value'),
             Input('facet_col', 'value'),
             Input('facet_row', 'value'),
             Input('sql-query-button', 'children'),
             ]
        )
        def multi_level(x_axis, y_axis, size, color, facet_col, facet_row, sql_query):
            graph_type = 'filtered_df'
            df3 = self.caching_data_manager(df, sql_query, graph_type)
            return px.scatter(df3, x=x_axis, y=y_axis, size=size, color=color, facet_col=facet_col, facet_row=facet_row,
                              facet_col_wrap=4)

           

        # Port Finder
        port = 8080
        debug_value= False


        if mode == "inline":
            try:
                app.run_server(mode="inline", port=port,debug=debug_value,dev_tools_ui=debug_value,dev_tools_props_check=debug_value)
            except:
                port= self.find_free_port()
                app.run_server(mode="inline",port=port,debug=debug_value,dev_tools_ui=debug_value,dev_tools_props_check=debug_value)
        else:
            try:
                app.run_server(host='0.0.0.0', port=port,debug=debug_value,dev_tools_ui=debug_value,dev_tools_props_check=debug_value)
            except:
                # try different ip in case 0.0.0.0 does not work
                try:
                    try:
                        port=self.find_free_port()
                        app.run_server(host='0.0.0.0', port=port,debug=debug_value,dev_tools_ui=debug_value,dev_tools_props_check=debug_value)
                    except:
                        port = self.find_free_port()
                        app.run_server(host='0.0.0.0', port=port,debug=debug_value,dev_tools_ui=debug_value,dev_tools_props_check=debug_value)
                except:
                    try:
                        port = self.find_free_port()
                        app.run_server(host='127.0.0.1', port=port,debug=debug_value,dev_tools_ui=debug_value,dev_tools_props_check=debug_value)
                    except:
                        print("Please restart Jupyter Notebook or Python IDE.")
                        return False
        try:
            self.increate_counter()
        except:
            pass
        return True


    def find_free_port(self):
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(('', 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]

    #COUNTER FUNCTION (NEEDS TO BE IMPLEMENTED)
    def increate_counter(self):
        # call api call here

        url = 'https://us-central1-explainx-25b88.cloudfunctions.net/increaseCounter'
        params = {

        }
        data = {
            "user_id": self.user_id,
            "model": self.param["model_name"]

        }
        r = requests.post(url, params=params, json=data)
        if r.json()["message"]=="200":
            return True
        else:
            return False



    def calculate_prediction_shap(self, df):
        if self.param["model_name"] == "xgboost":
            import xgboost
            if xgboost.__version__ in ['1.1.0', '1.1.1', '1.1.0rc2', '1.1.0rc1']:
                print(
                    "Current Xgboost version is not supported. Please install Xgboost using 'pip install xgboost==1.0.2'")
                return False
            prediction_col = self.param["model"].predict(xgboost.DMatrix(df))

        elif self.param["model_name"]  == "catboost":
            prediction_col = self.param["model"].predict(df.to_numpy())

        else:
            prediction_col = self.param["model"].predict(df.to_numpy())

        # is classification?
        is_classification = self.param["is_classification"]

        # shap
        c = calculate_shap()
        df_final, explainer = c.find(self.param["model"], df, prediction_col, is_classification, model_name=self.param["model_name"])

        # prediction col
        df_final[self.param["y_variable_predict"]] = prediction_col

        return df_final

    def calculate_prediction(self, df):
        if self.param["model_name"] == "xgboost":
            import xgboost
            if xgboost.__version__ in ['1.1.0', '1.1.1', '1.1.0rc2', '1.1.0rc1']:
                print(
                    "Current Xgboost version is not supported. Please install Xgboost using 'pip install xgboost==1.0.2'")
                return False
            prediction_col = self.param["model"].predict(xgboost.DMatrix(df))

        elif self.param["model_name"]  == "catboost":
            prediction_col = self.param["model"].predict(df.to_numpy())

        else:
            prediction_col = self.param["model"].predict(df.to_numpy())

        # is classification?
        is_classification = self.param["is_classification"]


        # prediction col
        df[self.param["y_variable_predict"]] = prediction_col

        return df



def localtunnel(port):
    # subdomain= 'explainx-'+ get_random_string(10)
    task = subprocess.Popen(['lt', '-h', '"https://serverless.social"', '-p', str(port)],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE, preexec_fn=os.setsid)

    outpt= task.stdout.readline()
    outpt_string= outpt.decode("utf-8").split("is:")
    print('Explainx.ai is running @ '+outpt_string[1])


def get_random_string(length):
    letters = string.ascii_lowercase+ string.ascii_uppercase
    result_str = ''.join(random.choice(letters) for i in range(length))
    return result_str




from imports import *

"""
This class calculates feature importance

Input: 


"""


class shap_pdp():
    def __init__(self):
        super(shap_pdp, self).__init__()
        self.param= None

    def find(self,  df):

        return df





from imports import *
from feature_impact import *
from feature_importance import *
from shap_pdp import *
from summary_plot import *
import dash_daq as daq
from data_for_shap_graphs import *

slider_feature = {"fontSize":"10px",
                    'white-space': 'normal',
                        "overflow-wrap": "break-word",
                        "width":"200px",
                        "padding":"1px", 
                        "border-bottom": "1px solid #2d3436",
                        'justify-content':'center',
                        'padding-top':'20px'}

class what_if():
    def __init__(self):
        super(what_if, self).__init__()
        self.data = data_for_shap_graphs()

        # save all important variables here.
    def col_values_pairs(self, df, features):
        df_features= df[features]
        df_describe= df_features.describe()
        
        """
        which features are categorical?
        find their feature value pairs
        """

        numnerical_col= df_describe.columns
        categorical_col= list(set(features)- set(numnerical_col))

        col_type_dict={}
        col_value_dict={}
        for f in numnerical_col:
            col_type_dict[f]="slider"
            col_value_dict[f]= self.get_min_max_featurs(df_describe, f)
            
        
        for f in categorical_col:
            categ,col_type= all_categories(df, f)
                
            col_type_dict[f]= col_type
            col_value_dict[f]= categ
            
        return col_value_dict, col_type_dict


    def all_categories(self, df, feature_name):
        
        """
        find all categories in categorical features.
        """
        
        
        all_categories=list(set(df[feature_name]))
        
    #     if len(all_categories)<=3:
    #         col_type= 'radio'
    #     else:
    #         col_type= 'dropdown'
        
        return all_categories, 'dropdown'

        
    def get_min_max_featurs(self, df_describe, feature_name):

        x_des_var_list=list(df_describe[feature_name])
        q00= x_des_var_list[3]-25/100*x_des_var_list[3]
        q0= x_des_var_list[3]
        q1= x_des_var_list[4]
        q2= x_des_var_list[5]
        q3= x_des_var_list[6]
        q4= x_des_var_list[7]
        q5= x_des_var_list[7]+ 25/100*x_des_var_list[7]
        ans= [q00,q0,q1,q2,q3,q4,q5]
        
        # round off to 2 dp
        for i in range(len(ans)):
            ans[i]= round(ans[i],2)
        
        
        return ans

    def what_if_form(self, df, features):
        col_value_dict, col_type_dict = self.col_values_pairs(df, features)
        
        form_group_array=[]
        for f in features:
            """
            if feature type== slider
            then following is correct.
            """
            if col_type_dict[f]=='slider':
                #form_group_array.append(self.form_group_slider(f, col_value_dict[f]))
                 form_group_array.append(self.form_group_input(f, col_value_dict[f]))
                
            elif col_type_dict[f]=='radio':
                form_group_array.append(self.form_group_radio(f, col_value_dict[f]))
            else:
                form_group_array.append(self.form_group_dropdown(f, col_value_dict[f]))
                                    
                                    
        return html.Form(form_group_array)
            
    def form_group_input(self, feature, values, value=None):
        marks= {}
        for v in values:
            marks[v]=str(v)
        
        if value==None:
            value=values[3]
        
        fg= dbc.FormGroup([
                    html.Div(feature, style = slider_feature),
                    html.Div(dcc.Input(
                        id=feature+'_slider',
                        type="number",
                        value = value,
                        debounce=True,
                        placeholder="Enter the value"), 
                            style={"font-size":"10px", 
                                    'width': "200px", 
                                    'display': 'inline-block', 
                                    "padding":"1px",
                                    'margin-left':"30px",
                                    "height":'40px'})

                        ], style={"display":"flex", "width":"450px","height":"40px"})
        
        return fg
        
        
    def form_group_slider(self, feature, values, value=None):
        marks= {}
        for v in values:
            marks[v]=str(v)
        
        if value==None:
            value=values[3]
        
        fg= dbc.FormGroup([
                    html.Div(feature, style = slider_feature),
                    html.Div(daq.Slider(
                                id=feature+'-slider',
                                min=values[0],
                                max=values[-1],
                            handleLabel={"showCurrentValue": True,"label": 'Value'},
                            ),style={'width': "100px", 
                                    'display': 'inline-block', 
                                    "padding":"1px",
                                    'margin-left':"20px",
                                    'margin-top':'25px',
                                    "height":'40px'})

                        ], style={"display":"flex", 
                                    "width":"400px",
                                    "height":"40px",
                                    'margin-bottom':'5px',
                                    'margin-top':'5px'})
        
        return fg


    def form_group_dropdown(self, feature, values, value=None):
        options= []
        for v in values:
            options.append({'label': v, 'value': v})
        
        if value==None:
            value=values[0]
        
        fg = dbc.FormGroup([
                    html.Div(feature,style = slider_feature),
                    html.Div(dcc.Dropdown(
                                id= feature+'-slider',
                                options=options,
                                value=value
                    ), 
                                style={"font-size":"9px", 
                                        'width': "200px", 
                                    'display': 'inline-block', 
                                    "padding":"1px",
                                    'margin-left':"30px",
                                    "height":'40px'})
        ],style={"display":"flex",  "width":"450px", "height":'40px'})
        
        return fg




from imports import *
from rescale_numeric_feature import *

"""
This class calculates feature importance

Input: 


"""


class calculate_shap():
    def __init__(self):
        super(calculate_shap, self).__init__()
        self.param = None

    def xgboost_shap(self, model, X):
        # explain the model's predictions using SHAP
        # (same syntax works for LightGBM, CatBoost, scikit-learn and spark models)
        explainer = shap.TreeExplainer(model)
        shap_values = explainer.shap_values(X)

        pd_shap = pd.DataFrame(shap_values)
        all_columns = list(X.columns)

        shap_columns = []

        for i in all_columns:
            shap_columns.append(i + "_impact")

        pd_shap.columns = shap_columns

        Y = X.copy()
        for c in shap_columns:
            Y[c] = list(pd_shap[c])

        return Y, explainer

    def catboost_shap(self, model, df, y_variable=None):
        import catboost
        from catboost import CatBoostClassifier, Pool
        from catboost import CatBoostRegressor
        from catboost.utils import get_confusion_matrix
        # explain the model's predictions using SHAP
        if y_variable != None:
            try:
                df = df.drop(y_variable, axis=1)
            except:
                df = df

        # find categorical variables and it's index
        g = get_cols()

        cat, cat_index = g.cate_col_with_index(df)
        all_columns = list(df.columns)

        # convert dataframe in to array
        df_array = df.to_numpy()

        # call the function
        shap_values = self.get_shap_values(df_array, model, all_columns, cat_index)

        # append the results with the original file

        # final_df = df.join(shap_values)

        shap_columns= shap_values.columns

        Y = df.copy()
        for c in shap_columns:
            Y[c] = list(shap_values[c])


        return Y


    def kernel_shap(self, model, X_train):
        # use Kernel SHAP to explain test set predictions
        explainer = shap.KernelExplainer(model.predict_proba, X_train)
        shap_values = explainer.shap_values(X_train, nsamples=100)

        pd_shap = pd.DataFrame(shap_values)
        all_columns = list(X_train.columns)

        shap_columns = []

        for i in all_columns:
            shap_columns.append(i + "_impact")
        pd_shap.columns = shap_columns



        Y = X_train.copy()
        for c in shap_columns:
            Y[c] = list(pd_shap[c])

        return Y, explainer

    def kernel_shap_classification(self, model, X_train,prediction_col):
        # use Kernel SHAP to explain test set predictions
        explainer = shap.KernelExplainer(model.predict_proba, X_train)
        shap_values = explainer.shap_values(X_train, nsamples=100)

        pd_shap = self.select_row_shap_values(shap_values, prediction_col)
        all_columns = list(X_train.columns)

        shap_columns = []

        for i in all_columns:
            shap_columns.append(i + "_impact")
        pd_shap.columns = shap_columns



        Y = X_train.copy()
        for c in shap_columns:
            Y[c] = list(pd_shap[c])

        return Y, explainer

    def select_row_shap_values(self, shap_values,prediction_col):

        num_of_classes = len(shap_values)

        if num_of_classes== len(prediction_col):
            df_final = pd.DataFrame(shap_values)
            return df_final

        point_no=0
        df_array = []
        for p in prediction_col:
            df_array.append(shap_values[p][point_no])
            point_no=point_no+1

        df_final = pd.DataFrame(df_array)
        return df_final


    def randomforest_shap_classification(self, model, X,prediction_col):
        explainer = shap.TreeExplainer(model)
        shap_values = explainer.shap_values(X,approximate=True)


        pd_shap = self.select_row_shap_values(shap_values,prediction_col)
        all_columns = list(X.columns)


        pd_shap.columns = [f"{y}_impact" for y in all_columns]

        shap_columns = pd_shap.columns

        Y = X.copy()
        for c in shap_columns:
            Y[c] = list(pd_shap[c])


        return Y, explainer


    def randomforest_shap(self, model, X):
        explainer = shap.TreeExplainer(model)
        shap_values = explainer.shap_values(X,approximate=True)


        pd_shap = pd.DataFrame(shap_values)
        all_columns = list(X.columns)


        pd_shap.columns = [f"{y}_impact" for y in all_columns]

        shap_columns = pd_shap.columns

        Y = X.copy()
        for c in shap_columns:
            Y[c] = list(pd_shap[c])


        return Y, explainer


    def get_shap_values(self, x_array, model, x_variable, cat_index):
        """
        SHAP VALUES CALCULATED
        """
        shap_values = model.get_feature_importance(Pool(x_array, cat_features=cat_index), type='ShapValues')
        shap_values = shap_values[:, :-1]
        total_columns = x_variable
        total_columns = [i + '_impact' for i in total_columns]

        shap_values = pd.DataFrame(data=shap_values, columns=total_columns)
        return shap_values

    def find(self, model, df,prediction_col,is_classification, model_name="xgboost"):

        if model_name == "xgboost":
            df2 , explainer= self.xgboost_shap(model, df)
            return df2, explainer

        elif model_name == "lightgbm":
            df2 , explainer= self.xgboost_shap(model, df)

            return df2, explainer

        elif model_name == "catboost":
            df2 = self.catboost_shap(model, df)
            explainer= None
            return df2, explainer


        elif model_name == "randomforest":
            if is_classification:
                df2 , explainer= self.randomforest_shap_classification(model, df, prediction_col)
            else:
                df2 , explainer= self.randomforest_shap(model, df)
            return df2, explainer

        elif model_name == "svm":
            if is_classification:
                df2 , explainer= self.kernel_shap_classification(model, df,prediction_col)
            else:
                df2, explainer = self.kernel_shap(model, df)
            return df2, explainer

        elif model_name == "knn":
            if is_classification:
                df2, explainer = self.kernel_shap_classification(model, df,prediction_col)
            else:
                df2 , explainer= self.kernel_shap(model, df)
            return df2, explainer

        elif model_name == "logisticregression":
            if is_classification:
                df2 , explainer= self.kernel_shap_classification(model, df,prediction_col)
            else:
                df2, explainer = self.kernel_shap(model, df)
            return df2, explainer

        elif model_name == "decisiontree":
            if is_classification:
                df2, explainer = self.kernel_shap_classification(model, df,prediction_col)
            else:
                df2, explainer = self.kernel_shap(model, df)
            return df2, explainer

        elif model_name == "neuralnetwork":
            if is_classification:
                df2, explainer = self.kernel_shap_classification(model, df,prediction_col)
            else:
                df2, explainer = self.kernel_shap(model, df)
            return df2, explainer

        elif model_name=="gradientboostingregressor":
            df2 , explainer= self.xgboost_shap(model, df)
            return df2, explainer
        elif "gradientboosting" in model_name:
            df2, explainer = self.xgboost_shap(model, df)
            return df2, explainer
        else:
            if is_classification:
                df2 , explainer= self.kernel_shap_classification(model, df,prediction_col)
            else:
                df2 , explainer= self.kernel_shap(model, df)
            return df2, explainer







from explainx import *

from catboost import CatBoostRegressor
import pandas as pd
from xgboost import XGBClassifier as xgb
import xgboost
from sklearn.model_selection import train_test_split

# train XGBoost model
X,y = explainx.dataset_heloc()


#xgboost
model = xgboost.train({"learning_rate": 0.01}, xgboost.DMatrix(X, label=y), 100)

explainx.ai(X, y, model, model_name="xgboost")

from sklearn.ensemble import GradientBoostingRegressor

# Load boston dataset
X,y = explainx.dataset_boston()

# split data into train and test.
X_train, X_test, y_train, y_test = train_test_split(X,y, test_size=0.2, random_state=0)

# GradientBoostingRegressor

model = GradientBoostingRegressor()
model.fit(X_train, y_train)

explainx.ai(X_test,  y_test, model, model_name="gradientboostingregressor")



# Load Heloc dataset
X,y = explainx.dataset_heloc()

# Split data into train and test.
X_train, X_test, y_train, y_test = train_test_split(X,y, test_size=0.2, random_state=0)


# Run catboost model
model = CatBoostRegressor(iterations=150,
                          learning_rate=.3,
                          depth=2)


# Fit model
model.fit(X_train.to_numpy(), y_train)

explainx.ai(X_test,  y_test, model, model_name="catboost")



from imports import *

"""
This class calculates feature importance

Input: 


"""


class insight_classification():
    def __init__(self):
        super(insight_classification, self).__init__()
        self.param = None

    def insight_1_feature_imp(self, df, classes=[0, 1]):
        #
        top_important_variables = []
        average_shap_values = []

        df_numpy = df.to_numpy()

        for i in range(3):
            top_important_variables.append(df_numpy[-1 * i][0])
            average_shap_values.append(df_numpy[-1 * i][1])

        sentences = []
        # sentences.append(
        #     "This graph helps you identify which features in your dataset have the greatest effect on the outcomes of your machine learning model.")
        sentences.append("In this graph, you can see: " + str(top_important_variables[0]) + " " +
                         str(top_important_variables[1]) + " " +
                         str(top_important_variables[
                                 2]) + " " + "are the top three most important variables according to your machine learning model.")

        sentences.append("Remember, each variable might affect the outcome differently.")

        sentences.append("For example, On average, the variable " + str(top_important_variables[0]) + " " +
                         "will change the probability of predicting any outcome by " +
                         str(round(average_shap_values[0] * 100, 2)) + "%.")

        # append all outcomes
        if len(classes) == 2:
            outcomes = str(classes[0]) + " and " + str(classes[1])
        else:
            outcomes = ""
            for c in classes[:-1]:
                outcomes = outcomes + str(c) + ", "
            outcomes = outcomes + "and " + str(classes[-1])

        sentences.append("But variable " + str(top_important_variables[0]) + " " +
                         "might affect the outcome " + outcomes +
                         " differently.")

        # sentences.append("Use our Filter Function to see how different variables affect the outcome " + str(classes[0]))

        sentences.append(
            "So let’s scroll down and explore how each variable impacts the outcome on average.")

        return sentences

    def insight_2_global_feature_impact(self, df, outcome, expected_values, classes=[0, 1]):

        top_positive_variables = []
        average_shap_values_positive = []

        top_negative_variables = []
        average_shap_values_negative = []

        #df_numpy = df.to_numpy()

        for i in range(1,4):
            top_positive_variables.append(df.iloc[-1 * i][0])
            average_shap_values_positive.append(df.iloc[-1 * i][1])

            top_negative_variables.append(df.iloc[i][0])
            average_shap_values_negative.append(df.iloc[i][1])

        sentences = []

        sentences.append("The average probability of predicting " +
                         str(outcome) + " is " + str(round(expected_values[outcome],2)) + ".\n")

        sentences.append("\nOn average, the variable " +
                         top_positive_variables[0] + ", " +
                         top_positive_variables[1] + ",and " +
                         top_positive_variables[2] + " " +
                         "will increase the probability for achieving outcome = " +
                         str(outcome) 
                         + " by " +
                         str(round(average_shap_values_positive[0] * 100, 2)) + "%," +
                         str(round(average_shap_values_positive[1] * 100, 2)) + "%, and " +
                         str(round(average_shap_values_positive[2] * 100, 2)) + "% respectively.\n")

        sentences.append("On average, the variable " +
                         top_negative_variables[0] + ", " +
                         top_negative_variables[1] + ",and " +
                         top_negative_variables[2] + " " +
                         "will change the probability for achieving outcome = " +
                         str(outcome) 
                         + " by " +
                         str(round(average_shap_values_negative[0] * 100, 2)) + "%," +
                         str(round(average_shap_values_negative[1] * 100, 2)) + "%, and" +
                         str(round(average_shap_values_negative[2] * 100, 2)) + "% respectively.")

        sentences.append("")

        return sentences

    def insight_3(self, df):

        top_positive_variables = []
        average_shap_values_positive = []

        df_numpy = df.to_numpy()

        for i in range(3):
            top_positive_variables.append(df_numpy[-1 * i][0])
            average_shap_values_positive.append(df_numpy[-1 * i][1])

        sentences = []

        sentences.append("Please explore the PDP graph to identify how different values of " +
                         top_positive_variables[0] + ", " +
                         top_positive_variables[1] + ", and " +
                         top_positive_variables[2] + " " +
                         " affect the model decision.")

        return sentences

    def insight_1_feature_imp(self, df, classes=[0, 1]):
        #
        top_important_variables = []
        average_shap_values = []

        df_numpy = df.to_numpy()

        for i in range(1,4):
            top_important_variables.append(df_numpy[-1 * i][0])
            average_shap_values.append(df_numpy[-1 * i][1])

        sentences = []
        # sentences.append(
        #     "This graph helps you identify which features in your dataset have the greatest effect on the outcomes of your machine learning model.")
        sentences.append("In this graph, you can see: " + str(top_important_variables[0]) + " " +
                         str(top_important_variables[1]) + " " +
                         str(top_important_variables[
                                 2]) + " " + "are the top three most important variables according to your machine learning model.")

        sentences.append("Remember, each variable might affect the outcome differently.")

        sentences.append("For example, On average, the variable " + str(top_important_variables[0]) + " " +
                         "will change the probability of predicting any outcome by " +
                         str(round(average_shap_values[0] * 100, 2)) + "%.")

        # append all outcomes
        if len(classes) == 2:
            outcomes = str(classes[0]) + " and " + str(classes[1])
        else:
            outcomes = ""
            for c in classes[:-1]:
                outcomes = outcomes + str(c) + ", "
            outcomes = outcomes + "and " + str(classes[-1])

        sentences.append("But variable " + str(top_important_variables[0]) + " " +
                         "might affect the outcome " + outcomes +
                         " differently.")

        # sentences.append("Use our Filter Function to see how different variables affect the outcome " + str(classes[0]))

        sentences.append(
            "So let’s scroll down and explore how each variable impacts the outcome on average.")

        return sentences

    def insight_2_local_feature_impact(self, df, outcome, expected_values, classes=[0, 1]):

        top_positive_variables = []
        average_shap_values_positive = []

        top_negative_variables = []
        average_shap_values_negative = []

        #df_numpy = df.to_numpy()

        for i in range(1,4):
            top_positive_variables.append(df.iloc[-1 * i][0])
            average_shap_values_positive.append(df.iloc[-1 * i][1])

            top_negative_variables.append(df.iloc[i][0])
            average_shap_values_negative.append(df.iloc[i][1])

        sentences = []

        sentences.append("The average probability of predicting " +
                         str(outcome) + " is " + str(round(expected_values[outcome],2)) + ".\n")

        sentences.append("\nOn average, the variable " +
                         top_positive_variables[0] + ", " +
                         top_positive_variables[1] + ",and " +
                         top_positive_variables[2] + " " +
                         "will increase the probability for achieving outcome = " +
                         str(outcome) 
                         + " by " +
                         str(round(average_shap_values_positive[0] * 100, 2)) + "%, " +
                         str(round(average_shap_values_positive[1] * 100, 2)) + "%, and " +
                         str(round(average_shap_values_positive[2] * 100, 2)) + "% respectively.\n")

        sentences.append("On average, the variable " +
                         top_negative_variables[0] + ", " +
                         top_negative_variables[1] + ",and " +
                         top_negative_variables[2] + " " +
                         "will change the probability for achieving outcome = " +
                         str(outcome) 
                         + " by " +
                         str(round(average_shap_values_negative[0] * 100, 2)) + "%, " +
                         str(round(average_shap_values_negative[1] * 100, 2)) + "%, and " +
                         str(round(average_shap_values_negative[2] * 100, 2)) + "% respectively.")

        sentences.append("")

        return sentences

    def insight_4_pdp(self, df):
        output = True
        return output

    def insight_5(self, x):
        output = True
        return output






from imports import *
from plotly_graphs import *
from plotly_css import *
from protodash import *
from app import app
import pandas as pd

def layout_local(original_variables,columns,df_column):
    layout = html.Div([
            # Input and Data
            html.Div([
                html.Div([
                    html.H4('What-If Analysis',
                            style=style26),
                    html.I("Datapoint Index."),

                    dcc.Input(id="row_number", type="number", value=1,
                                placeholder="Enter a Row Number e.g. 1, 4, 5",
                                style={'text-align': 'center'}),
                    html.Button(id='submit-button-state', n_clicks=0, children='Predict', style={'margin-left':'50px'}),
                    

                    # html.Div(id='local_data_table', style={'marginTop': "20px", 'marginBottom': "5px"}),


                ]),

            ]),  # End of Input & Data Div

            html.Div([

                #What-If Div
                html.Div([

                    

                    html.Br(),
                    html.Div([
                        
                        dbc.Table(html.Thead(html.Tr([html.Th("Feature", style={"width":"50%"}), html.Th("Value")])), 
                        bordered=True, 
                        dark=True,
                        hover=True,
                        responsive=True,
                        striped=True,
                        style={'width':'88%'}),

                    html.Div(id="place_form_here", 
                        style={ 
                            "maxHeight": "650px", 
                            "maxWidth":"500px",
                            "overflow": "scroll",
                            # "border-right":"1px solid grey",
                            # "border-bottom":"1px solid grey",
                            # "border-top":"1px solid grey",
                            "border-radius":'0 0 10px 10px',
                            'box-shadow': '0 4px 8px 0 rgba(0, 0, 0, 0.2), 0 6px 20px 0 rgba(0, 0, 0, 0.19)',
                            "display":"flex",
                            "justify-content":"center",
                            "padding-top":"10px",
                            "padding-bottom":"10px"}),
                   
                      
                      ])

                ], style={'width':'35%'}),

                #Tabs Div
                html.Div([

                     dcc.Tabs([
                dcc.Tab(label='Local Feature Explanation', children=[

                    html.Div(id='datatable-interactivity-container2', children=[
                        html.Div([

                            html.Div([
                                html.H4('Local Feature Impact',
                                        style=style29),
                                html.P(
                                    'Local Feature impact identifies which features have the greatest positive or negative effect on the outcome of a machine learning model for a specific row.',
                                    style=style30),
                                dcc.Loading(
                                    id="local_feature_impact_1",
                                    type="circle",
                                    children=dbc.Row(
                                        [
                                            dbc.Col(html.Div(dcc.Graph(id='local_feature_impact',
                                                                        style={'marginLeft': 75,
                                                                                'marginTop': 0,
                                                                                'height': '500px'})),
                                                    width=8),
                                            dbc.Col(
                                                [
                                                    html.Div([
                                                        html.H2("How to read this graph?"),
                                                        html.P(
                                                            "According to the model, the features are most important in explaining the target variable. Most importance is on the top.")
                                                    ]),
                                                    html.Div([
                                                        html.H2("Insights"),
                                                        html.P(id='local_message_1'),
                                                        html.P(id='local_message_2'),
                                                        html.P(id='local_message_3')
                                                    ]),
                                                    html.Div([
                                                        html.H2("Next Steps"),
                                                        html.P(
                                                            "Click on the similar profile tabs and explore profiles that had similar attributes"),
                                                    ])
                                                ]
                                                , width=4),

                                        ]))

                            ],
                                style=style31,
                            ),
                        ],

                            style=style32)], style={'height': '400'})

                ]),

                dcc.Tab(id="similar_tabs", label='Similar Profiles', children=[

                    html.Div([
                        html.H4('Similar Profiles',
                                style=style33),

                        html.Button('Display Similar Profiles', id='btn-nclicks-1', n_clicks=0,
                                    style=style34),

                        html.P(
                            'Display similar  profiles and the extent to which they are similar to the chosen applicant as indicated by the last row in the table below labelled as "Weight"',
                            style=style35),
                        html.Div([
                            dcc.Loading(
                                id="table",
                                type="circle",
                                children=dash_table.DataTable(
                                    id='prototype_data',
                                    columns=[{"name": i, "id": i} for i in columns],
                                    data=[],
                                    editable=False,
                                    # filter_action="native",
                                    sort_mode="multi",
                                    row_deletable=False,
                                    page_action="native",
                                    page_current=0,
                                    page_size=len(original_variables) + 1,
                                    style_table=style36,
                                    style_header=style37,
                                    style_cell={
                                        "font-family": "Helvetica, Arial, sans-serif",
                                        "fontSize": "15px",
                                        # 'width': '{}%'.format(len(df.columns)),
                                        'width': '{}%'.format(len(df_column)),
                                        'textOverflow': 'ellipsis',
                                        'overflow': 'hidden',
                                        'textAlign': 'left',
                                        'minWidth': '180px', 'width': '180px', 'maxWidth': '180px',
                                        "marginBottom": "100px"
                                    },
                                    style_data_conditional=[
                                        {
                                            'if': {

                                                'filter_query': '{index} = "Weight(%)"'

                                            },
                                            "backgroundColor": "green",
                                            'color': 'white'

                                        },

                                        {

                                            'if': {

                                                'filter_query': '{0_color} =1',

                                                'column_id': '0'

                                            },

                                            'backgroundColor': 'tomato',

                                            'color': 'white'

                                        },

                                        {

                                            'if': {

                                                'filter_query': '{1_color} =1',

                                                'column_id': '1'

                                            },

                                            'backgroundColor': 'tomato',

                                            'color': 'white'

                                        },

                                        {

                                            'if': {

                                                'filter_query': '{2_color} =1',

                                                'column_id': '2'

                                            },

                                            'backgroundColor': 'tomato',

                                            'color': 'white'

                                        },

                                        {

                                            'if': {

                                                'filter_query': '{3_color} =1',

                                                'column_id': '3'

                                            },

                                            'backgroundColor': 'tomato',

                                            'color': 'white'

                                        },

                                        {

                                            'if': {

                                                'filter_query': '{4_color} =1',

                                                'column_id': '4'

                                            },

                                            'backgroundColor': 'tomato',

                                            'color': 'white'

                                        }

                                    ],
                                    
                                )
                            )

                        ])

                    ], style=style39),

                ])

                    ])

                ], style={"width":'65%'})

            ], style={'display':'flex'}),
        
             ])
    

    return layout







from imports import *
from plotly_graphs import *
from plotly_css import *
from app import app


def layout_distribution(original_variables):
    layout = html.Div([
            html.Div([
                html.Div([
                    html.H4('Distributions',
                            style=style21),
                    html.Div([
                        html.P('Variable Name'),
                        dcc.Dropdown(
                            id='xaxis-column-name',
                            options=[{'label': i, 'value': i} for i in original_variables],
                            value= original_variables[0]
                        ),
                    ],
                        style=style22),
                    html.Div([
                        html.P('Plot Type'),
                        dcc.Dropdown(
                            id='plot_type',
                            options=[{'label': 'Violin Plot', 'value': 'Violin Plot'},
                                        {'label': 'Histogram', 'value': 'Histogram'}, ],
                            value='Histogram'
                        ),
                    ],
                        style=style23),

                ]),

                dcc.Graph(id='indicator-graphic2',
                            style={'marginLeft': 50, 'marginTop': 100, 'height': '500px'}),
            ],
                style=style24),

            html.Div([

                html.Div([
                    html.H4('Multi-Level EDA', style=style21),
                    html.Div([
                        html.P('X-Axis Variable'),
                        dcc.Dropdown(
                            id='x_axis',
                            options=[{'label': i, 'value': i} for i in original_variables],
                            value=original_variables[0]
                        ),
                    ], style=style22),
                    html.Div([
                        html.P('Y-Axis Variable'),
                        dcc.Dropdown(
                            id='y_axis',
                            options=[{'label': i, 'value': i} for i in original_variables],
                            value=original_variables[1]
                        ),
                    ], style=style22),
                    html.Div([
                        html.P('Size'),
                        dcc.Dropdown(
                            id='size',
                            options=[{'label': i, 'value': i} for i in original_variables],
                            # value=original_variables[0]
                            placeholder="Choose a variable"
                        ),
                    ], style=style22),
                    html.Div([
                        html.P('Color'),
                        dcc.Dropdown(
                            id='color',
                            options=[{'label': i, 'value': i} for i in original_variables],
                            placeholder="Choose a variable"
                        ),
                    ], style=style22),
                    html.Div([
                        html.P('Drill-down by Column'),
                        dcc.Dropdown(
                            id='facet_col',
                            options=[{'label': i, 'value': i} for i in original_variables],
                            placeholder="Select a categorical variable"
                            # value=original_variables[0]
                        ),
                    ], style=style22),
                    html.Div([
                        html.P('Drill-down by Row'),
                        dcc.Dropdown(
                            id='facet_row',
                            options=[{'label': i, 'value': i} for i in original_variables],
                            placeholder="Select a categorical variable"
                        ),
                    ], style=style22)
                ]),
                dcc.Graph(id='multi_level_eda',
                            style={'marginLeft': 50, 'marginTop': 200, 'height': '700px'})

            ], style=style24)
        ], style=style25)
    return layout



from imports import *

"""
This class calculates feature importance

Input: 


"""


class get_cols():
    def __init__(self):
        super(get_cols, self).__init__()
        self.param = None

    def get_all_cols(self, df):
        all_columns = df.columns

        original_columns = []
        for c in all_columns:
            if "_impact" in c:
                pass
            elif c + "_impact" in all_columns:
                original_columns.append(c)

        return original_columns

    def get_cate_numer_col(self, df):

        original_columns = self.get_all_cols(df)

        categorical_columns = self.cate_col(df)
        numeric_columns = set(original_columns) - set(categorical_columns)
        numeric_columns = list(numeric_columns)
        return numeric_columns, categorical_columns

    def cate_col(self, df):
        df= df.convert_dtypes()
        di = dict(df.dtypes)
        c = df.columns

        sample_df = pd.DataFrame(["a", "b", "c"])
        sample_df = sample_df.convert_dtypes()
        sample_df_dic= dict(sample_df.dtypes)

        cate = []
        for i in c:
            try:
                if di[i] == sample_df_dic[0]:
                    cate.append(i)
            except:
                pass

        return cate

    def cate_col_with_index(self, df):
        df = df.convert_dtypes()
        di = dict(df.dtypes)
        c = df.columns

        sample_df = pd.DataFrame(["a", "b", "c"])
        sample_df = sample_df.convert_dtypes()
        sample_df_dic = dict(sample_df.dtypes)

        cate = []
        index=[]
        j=0
        for i in c:
            try:
                if di[i] == sample_df_dic[0]:
                    cate.append(i)
                    index.append(j)
            except:
                pass
            j = j + 1


        return cate, index


class rescale_numeric_features():
    def __init__(self):
        super(rescale_numeric_features, self).__init__()
        self.param = None

    def get_min_max(self, df_describe, variable_name):
        mini = df_describe[variable_name][3]
        maxi = df_describe[variable_name][-1]
        return mini, maxi

    def add_col_rescaled(self, df):

        df_describe = df.describe()

        # let's find numeric and categorical columns
        column = get_cols()
        numeric_columns, categorical_columns = column.get_cate_numer_col(df)

        for nc in numeric_columns:
            # get min and max
            mini, maxi = self.get_min_max(df_describe, nc)

            df[nc + "_rescaled"] = (df[nc] - mini) / (maxi - mini) * 10

        for cc in categorical_columns:
            df[cc + "_rescaled"] = 0

        return df

    def rescale(self, df):
        column = get_cols()
        original_columns = column.get_all_cols(df)

        df_with_rescaled_features = self.add_col_rescaled(df)

        return df_with_rescaled_features









from imports import *
from sklearn.preprocessing import OneHotEncoder
import numpy as np
"""
This class calculates feature importance

Input: 


"""


class encode_decode_cat_col():
    def __init__(self):
        super(encode_decode_cat_col, self).__init__()
        self.param = None
        self.column_names = {}
        self.column_names_array = {}
        self.one_hot_encoder = {}

    def encode_col(self, df, column_name):
        if column_name in self.column_names.values():
            df = self.encode_existing(df, column_name)
        else:
            df = self.encode_new(df, column_name)

        return df

    def encode_new(self, df, column_name):
        self.one_hot_encoder[column_name] = OneHotEncoder(handle_unknown='ignore')

        # fill missing values with the a new category missing
        df[column_name] = df[column_name].fillna("missing")

        en = self.one_hot_encoder[column_name].fit_transform(df[[column_name]]).toarray()
        df_en = pd.DataFrame(en)

        # change column_names
        self.column_names[column_name], self.column_names_array[column_name] = self.change_column_names(df_en,
                                                                                                        column_name)
        df_en = df_en.rename(columns=self.column_names[column_name])

        # add new column and merge enc with the existing dataframe
        final_df = df.join(df_en)

        final_df = final_df.drop(column_name, axis=1)

        return final_df

    def encode_existing(self, df, column_name):

        # fill missing values with the a new category missing
        df[column_name] = df[column_name].fillna("missing")

        en = self.one_hot_encoder[column_name].fit_transform(df[[column_name]]).toarray()
        df_en = pd.DataFrame(en)

        # change column_names
        df_en = df_en.rename(columns=self.column_names[column_name])

        # add new column and merge enc with the existing dataframe
        final_df = df.join(df_en)

        final_df = final_df.drop(column_name, axis=1)

        return final_df

    def decode_col(self, df, column_name):

        # get columns
        df_en = df[self.column_names_array[column_name]]

        # get model
        arr = df_en.to_numpy()
        re_en = self.one_hot_encoder[column_name].inverse_transform(arr)

        df2 = pd.DataFrame(re_en)

        df2 = df2.rename(columns={0: column_name})

        # merge new col to the df
        final_df = df.join(df2)

        # remove existing columns
        final_df = final_df.drop(self.column_names_array[column_name], axis=1)

        return final_df

    def change_column_names(self, df, column_name):
        new_columns_names = {}
        col_array = []
        for i in list(df.columns):
            new_columns_names[i] = column_name + "_" + str(i)
            col_array.append(column_name + "_" + str(i))
        return new_columns_names, col_array








from imports import *

"""
This class calculates feature importance

Input: 


"""


class feature_importance():
    def __init__(self):
        super(feature_importance, self).__init__()
        self.param= None


    def find(self,  df):

        variables = [col for col in df.columns if '_impact' in col]
        y = []
        for i in range(len(variables)):
            p = df[variables[i]].abs().mean()
            y.append(p)
        res = {variables[i]: y[i] for i in range(len(y))}
        res2 = {k: v for k, v in sorted(res.items(), key=lambda item: item[1])}
        res3 = pd.Series(res2, name='Impact_Value')
        res3.index.name = 'VariableName'
        fi = res3.reset_index()
        return fi





import dash
from imports import *
from plotly_graphs import *
from plotly_css import *

external_stylesheets = ['https://codepen.io/rab657/pen/LYpKraq.css',
                            {
                                'href': 'https://stackpath.bootstrapcdn.com/bootstrap/4.5.0/css/bootstrap.min.css',
                                'rel': 'stylesheet'
                            }
                            ]

app = JupyterDash(__name__, external_stylesheets=external_stylesheets, suppress_callback_exceptions=True)

app.title = "explainX.ai - Main Dashboard"


server = app.server


style1 = {'backgroundColor': '#fff',
          'color': 'black',
          'border-radius': '15px 15px 0px 0px ',
          'textAlign': 'left',
          'height': "50px",
          'margin': 'auto',
          'padding-left': '70px',
          "font-family": "Helvetica, Arial, sans-serif", }

style2 = {'overflowX': 'auto',
          'margin': 'auto',
          "padding-left": '10px',
          'width': '90%'}

style3 = {
    "fontSize": "15px",
    'marginLeft': "10px",
    'color': 'black',
    'backgroundColor': 'rgb(230, 230, 230)',
    'fontWeight': 'bold'}

style4 = {'backgroundColor': "#fff", 'marginTop': 10, 'marginBottom': 10}

style5 = {'backgroundColor': '#fff',
          'color': 'black',
          'border-radius': '15px 15px 15px 15px ',
          'textAlign': 'left',
          'height': "50px",
          'margin': 'auto',
          'padding-left': '70px',
          'padding-top': '20px'

          }

style6 = {'backgroundColor': '#fff',
          'color': 'black',
          'border-radius': '15px 15px 0px 0px ',
          'textAlign': 'left',
          'height': "50px",
          'margin': 'auto',
          'padding-left': '70px',
          'padding-top': '20px',
          "font-family": "Helvetica, Arial, sans-serif",

          }

style7 = {'backgroundColor': "#fff", "minHeight": "400px",
          'border-radius': '15px 15px 15px 15px',
          'box-shadow': '0 4px 8px 0 rgba(0, 0, 0, 0.2), 0 6px 20px 0 rgba(0, 0, 0, 0.19)',
          'border-right': '1px solid #2c3e50', 'border-bottom': '1px solid #2c3e50'}

style8 = {'backgroundColor': '#fff',
          'color': 'black',
          'border-radius': '15px 15px 15px 15px ',
          'textAlign': 'left',
          'height': "50px",
          'margin': 'auto',
          'padding-left': '70px',
          'padding-top': '20px'

          }

style9 = {'backgroundColor': '#fff',
          'color': 'black',
          'border-radius': '15px 15px 0px 0px ',
          'textAlign': 'left',
          'height': "50px",
          'margin': 'auto',
          'padding-left': '70px',
          'padding-top': '20px',
          "font-family": "Helvetica, Arial, sans-serif",

          }

style10 = {'backgroundColor': "#fff", "minHeight": "400px",
           'border-radius': '15px 15px 15px 15px',
           'box-shadow': '0 4px 8px 0 rgba(0, 0, 0, 0.2), 0 6px 20px 0 rgba(0, 0, 0, 0.19)',
           'border-right': '1px solid #2c3e50', 'border-bottom': '1px solid #2c3e50',
           'marginTop': 50}

style11 = {
    "marginTop": 50,
    'border-radius': '15px 15px 15px 15px',

}

style12 = {'backgroundColor': '#fff',
           'color': 'black',
           'border-radius': '15px 15px 15px 15px ',
           'textAlign': 'left',
           'height': "50px",
           'margin': 'auto',
           'padding-left': '70px',
           'padding-top': '20px'

           }

style13 = {'backgroundColor': '#fff',
           'color': 'black',
           'border-radius': '15px 15px 0px 0px ',
           'textAlign': 'left',
           'height': "50px",
           'margin': 'auto',
           'padding-left': '70px',
           'padding-top': '20px',
           "font-family": "Helvetica, Arial, sans-serif",

           }

style14 = {'width': '20%', 'float': 'center', 'display': 'inline-block'}

style15 = {'width': '20%', 'marginRight': 10, 'float': 'center',
           'display': 'inline-block'}

style16 = {'backgroundColor': "#fff", "minHeight": "600px",
           'border-radius': '15px 15px 15px 15px',
           'box-shadow': '0 4px 8px 0 rgba(0, 0, 0, 0.2), 0 6px 20px 0 rgba(0, 0, 0, 0.19)',
           'border-right': '1px solid #2c3e50', 'border-bottom': '1px solid #2c3e50',
           'marginTop': 50, 'width': '95%'}

style17 = {'backgroundColor': '#fff',
           'color': 'black',
           'border-radius': '15px 15px 15px 15px ',
           'textAlign': 'left',
           'height': "50px",
           'margin': 'auto',
           'padding-left': '70px',
           'padding-top': '20px'

           }

style18 = {'backgroundColor': '#fff',
           'color': 'black',
           'border-radius': '15px 15px 0px 0px ',
           'textAlign': 'left',
           'height': "50px",
           'margin': 'auto',
           'padding-left': '70px',
           'padding-top': '20px',
           "font-family": "Helvetica, Arial, sans-serif",

           }

style19 = {'backgroundColor': "#fff", "minHeight": "600px",
           'border-radius': '15px 15px 15px 15px',
           'box-shadow': '0 4px 8px 0 rgba(0, 0, 0, 0.2), 0 6px 20px 0 rgba(0, 0, 0, 0.19)',
           'border-right': '1px solid #2c3e50', 'border-bottom': '1px solid #2c3e50',
           'marginTop': 50, }

style20 = {'marginBottom': 50,
           'marginTop': 50,
           'marginLeft': "1%",
           'width': '95%', }

style21 = {'backgroundColor': '#fff',
           'color': 'black',
           'border-radius': '15px 15px 15px 15px ',
           'textAlign': 'left',
           'height': "50px",
           'margin': 'auto',
           'padding-left': '70px',
           'padding-top': '20px'

           }

style22 = {'width': '25%', 'marginLeft': 70, 'float': 'left',
           'display': 'inline-block'}

style23 = {'width': '25%', 'marginLeft': 70, 'float': 'left',
           'display': 'inline-block'}

style24 = {'backgroundColor': "#fff", "minHeight": "700px",
           'border-radius': '15px 15px 15px 15px',
           'box-shadow': '0 4px 8px 0 rgba(0, 0, 0, 0.2), 0 6px 20px 0 rgba(0, 0, 0, 0.19)',
           'border-right': '1px solid #2c3e50', 'border-bottom': '1px solid #2c3e50',
           'marginTop': 50}

style25 = {'marginBottom': 50,
           'marginTop': 50,
           'marginLeft': "1%",
           'width': '95%', }

style26 = {'backgroundColor': '#fff',
           'color': 'black',
           'border-radius': '15px 15px 0px 0px ',
           'textAlign': 'left',
           'height': "50px",
           'margin': 'auto',
        #    'padding-left': '70px',
           'padding-top': '20px',
           "font-family": "Helvetica, Arial, sans-serif",

           }

style27 = {
    'backgroundColor': '#0984e3',
    'fontWeight': 'normal',
    "fontSize": "15px",
    'marginLeft': "10px",
    'color': 'white'
}

style28 = {'backgroundColor': "#fff", "minHeight": "200px",
           'border-radius': '15px 15px 15px 15px',
           'box-shadow': '0 4px 8px 0 rgba(0, 0, 0, 0.2), 0 6px 20px 0 rgba(0, 0, 0, 0.19)',
           'border-right': '1px solid #2c3e50', 'border-bottom': '1px solid #2c3e50',
           'marginTop': 50}

style29 = {'backgroundColor': '#fff',
           'color': 'black',
           'border-radius': '15px 15px 15px 15px ',
           'textAlign': 'left',
           'height': "50px",
           'margin': 'auto',
           'padding-left': '70px',
           'padding-top': '20px'

           }

style30 = {'backgroundColor': '#fff',
           'color': 'black',
           'border-radius': '15px 15px 0px 0px ',
           'textAlign': 'left',
           'height': "50px",
           'margin': 'auto',
           'padding-left': '70px',
           'padding-top': '20px',
           "font-family": "Helvetica, Arial, sans-serif",

           }

style31 = {'backgroundColor': "#fff", "minHeight": "400px",
           'border-radius': '15px 15px 15px 15px',
           'box-shadow': '0 4px 8px 0 rgba(0, 0, 0, 0.2), 0 6px 20px 0 rgba(0, 0, 0, 0.19)',
           'border-right': '1px solid #2c3e50',
           'border-bottom': '1px solid #2c3e50', 'marginTop': 50}

style32 = {
    "marginTop": 50,
    'border-radius': '15px 15px 15px 15px'}

style33 = {'backgroundColor': '#fff',
           'color': 'black',
           'border-radius': '15px 15px 0px 0px ',
           'textAlign': 'left',
           'height': "50px",
           'margin': 'auto',
           'padding-left': '70px',
           'padding-top': '20px',
           "font-family": "Helvetica, Arial, sans-serif",

           }

style34 = {'backgroundColor': '#fff',
           'color': 'black',
           'border-radius': '15px 15px 15px 15px ',
           'textAlign': 'left',
           'height': "50px",
           'margin': 'auto',
           'margin-left': '70px',
           "font-family": "Helvetica, Arial, sans-serif",

           }

style35 = {'backgroundColor': '#fff',
           'color': 'black',
           'border-radius': '15px 15px 0px 0px ',
           'textAlign': 'left',
           'height': "50px",
           'margin': 'auto',
           'padding-left': '70px',
           'padding-top': '20px',
           "font-family": "Helvetica, Arial, sans-serif",

           }

style36 = {'overflowX': 'scroll', 'margin': 'auto',
           "padding-left": '30px', 'width': '95%'}

style37 = {
    'backgroundColor': '#0984e3',
    'fontWeight': 'bold',
    "fontSize": "15px",
    'marginLeft': "10px",
    'textAlign': 'center',
    'color': 'white'
}


style39 = {'backgroundColor': "#fff", "minHeight": "800px",
           'border-radius': '15px 15px 15px 15px',
           'box-shadow': '0 4px 8px 0 rgba(0, 0, 0, 0.2), 0 6px 20px 0 rgba(0, 0, 0, 0.19)',
           'border-right': '1px solid #2c3e50', 'border-bottom': '1px solid #2c3e50',
           'marginTop': 50, 'width': '100%'}

style40 = {'marginBottom': 50,
           'marginTop': 10,
           'marginLeft': "1%",
           'marginRight': "1%",
           'width': '97%'}



import abc
import sys

# Ensure compatibility with Python 2/3
if sys.version_info >= (3, 4):
    ABC = abc.ABC
else:
    ABC = abc.ABCMeta(str('ABC'), (), {})


class DIExplainer(ABC):
    """
    DIExplainer is the base class for Directly Interpretable unsupervised explainers (DIE).
    Such explainers generally rely on unsupervised techniques to explain datasets and model predictions.
    Examples include DIP-VAE[#1]_, Protodash[#2]_, etc.

    References:
        .. [#1] Variational Inference of Disentangled Latent Concepts from Unlabeled Observations (DIP-VAE), ICLR 2018.
         Kumar, Sattigeri, Balakrishnan. https://arxiv.org/abs/1711.00848
        .. [#2] ProtoDash: Fast Interpretable Prototype Selection, 2017.
        Karthik S. Gurumoorthy, Amit Dhurandhar, Guillermo Cecchi.
        https://arxiv.org/abs/1707.01212
    """

    def __init__(self, *argv, **kwargs):
        """
        Initialize a DIExplainer object.
        ToDo: check common steps that need to be distilled here.
        """

    @abc.abstractmethod
    def set_params(self, *argv, **kwargs):
        """
        Set parameters for the explainer.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def explain(self, *argv, **kwargs):
        """
        Explain the data or model.
        """
        raise NotImplementedError



import os
from sys import platform


if platform == "linux" or platform == "linux2":
    try:
        os.system("sudo apt install nodejs")
        os.system("sudo apt install npm")
    except:
        os.system("sudo yum install nodejs")
        os.system("sudo yum install npm")
    os.system("npm install -g localtunnel")

elif platform == "darwin":
    os.system("xcode-select --install")
    os.system("brew install nodejs")
    os.system("npm install -g localtunnel")

elif platform == "win32":
    print("Please install nodejs, npm, and localtunnel manually")
    os.system("npm install -g localtunnel")
elif platform == "win64":
    print("Please install nodejs, npm, and localtunnel manually")
    os.system("npm install -g localtunnel")







class calculate_metrics():
    def __init__(self):
        super(calculate_metrics, self).__init__()
        self.param = None

    def classification_metrics(self, y_test, model_predict):
       
        # Acuracy_score
        acuracy_score = metrics.accuracy_score(y_test, model_predict) 

        # Log Loss
        y_probs = model.predict_proba(X_test)
        LogLoss = log_loss(y_test, y_probs, labels=model.classes_)

        # Explained Variance Score
        explained_variancescore = explained_variance_score(y_test, model_predict)


        fpr, tpr, thresholds =metrics.roc_curve(y_test, model_predict, pos_label=2)
        auc = metrics.auc(fpr, tpr)

        # MAE
        mae = mean_absolute_error(y_test, model_predict) 
        # MSE
        mse = mean_squared_error(y_test, model_predict)
        # RMS
        rms = sqrt(mse)   

        # Precision, Recall, F1-score, Support
        report = classification_report(y_test, model_predict,output_dict=True)
        report_dataframe = pd.DataFrame(report)
        report_dataframe = report_dataframe.transpose()

        # Confusion Matrix
        matrix = confusion_matrix(y_test, model_predict)
        matrix_dataframe = pd.DataFrame(matrix)

        # Make DataFrames
        metric = ["Accuracy Score",'Cross-Entropy Loss','Area Under Curve','MAE','MSE','RMS']
        values = [acuracy_score, LogLoss, auc, mae, mse, rms]
        metrics_dataframe = pd.DataFrame({'metric': metric, 'values': values})
        metrics_dataframe.set_index('metric', inplace = True)

    
   
        return metrics_dataframe, report_dataframe, matrix_dataframe
     
    
    
    
    def regression_metrics(self, y_test, model_predict):
       
        # #Explained variance regression score 
        exp_variance_score = explained_variance_score(y_test, model_predict)

        #max_error metric calculates the maximum residual error.
        maxerror = max_error(y_test, model_predict)

        #Mean absolute error regression loss
        mae = mean_absolute_error(y_test, model_predict) 

        #Mean squared error regression loss
        mse = mean_squared_error(y_test, model_predict) 

        # RMSE
        rmse = sqrt(mse)

        # R^2 (coefficient of determination) regression score function.
        r2 = r2_score(y_test, model_predict) 

        def mean_absolute_percentage_error(y_true, y_pred): 
            y_true, y_pred = np.array(y_true), np.array(y_pred)
            return np.mean(np.abs((y_true - y_pred) / y_true)) * 100

        mape = mean_absolute_percentage_error(y_test, model_predict)

        # Make DataFrames
        metric = ["Explained Variance Score",'Max Error','R squared','MAE','MSE','RMSE','MAPE']
        values = [exp_variance_score, maxerror, r2, mae, mse, rmse, mape]
        metrics_dataframe = pd.DataFrame({'metric': metric, 'values': values})
        metrics_dataframe.set_index('metric', inplace = True)

        return metrics_dataframe



from imports import *
from feature_impact import *
from feature_importance import *
from shap_pdp import *
from summary_plot import *

class data_for_shap_graphs():
    def __init__(self):
        super(data_for_shap_graphs, self).__init__()
        self.param= None

        # save all important variables here.

    def feature_importance(self,  df):
        fimp = feature_importance()
        df3 = fimp.find(df)
        return df3

    def feature_impact(self, df):
        fi = feature_impact()
        df2 = fi.find(df)
        return df2

    def summary_plot(self, df):
        sp = summary_plot()
        df5 = sp.find(df)
        return df5

    def partial_dependence_plot(self, df):
        pdp = shap_pdp()
        df4 = pdp.find(df)
        return df4



from imports import *
from feature_impact import *
from feature_importance import *
from shap_pdp import *
from summary_plot import *
from data_for_shap_graphs import *


class plotly_graphs():
    def __init__(self):
        super(plotly_graphs, self).__init__()
        self.data = data_for_shap_graphs()

        # save all important variables here.

    def feature_importance_old(self, df):
        df2 = self.data.feature_importance(df)

        names = list(df2["VariableName"])
        new_names = []

        for n in names:
            new_names.append(n.strip("_impact"))

        df2["VariableName"] = new_names

        feature_importance = px.bar(df2, x='Impact_Value', y="VariableName", orientation='h')
        return feature_importance, df2

    def feature_importance(self, df):
        df2 = self.data.feature_importance(df)

        names = list(df2["VariableName"])
        new_names = []

        for n in names:
            new_names.append(n.strip("_impact"))

        df2["VariableName"] = new_names

        feature_importance = px.bar(df2, x='Impact_Value', y="VariableName", orientation='h')
        # return feature_importance, df2
        return df2

    def feature_importance_graph(self, df):
        feature_importance = px.bar(df, x='Impact_Value', y="VariableName", orientation='h')
        return feature_importance

    def feature_impact_old(self, df):
        df2 = self.data.feature_impact(df)

        names = list(df2["VariableName"])
        new_names = []

        for n in names:
            new_names.append(n.strip("_impact"))

        df2["VariableName"] = new_names

        # Feature Impact Plot
        feature_impact = px.bar(df2, x='Impact_Value', y="VariableName", orientation='h')
        return feature_impact, df2

    def feature_impact(self, df):
        df2 = self.data.feature_impact(df)

        names = list(df2["VariableName"])
        new_names = []

        for n in names:
            new_names.append(n.strip("_impact"))

        df2["VariableName"] = new_names

        # Feature Impact Plot
        feature_impact = px.bar(df2, x='Impact_Value', y="VariableName", orientation='h')
        # return feature_impact, df2
        return df2

    def feature_impact_graph(self, df):
        feature_impact = px.bar(df, x='Impact_Value', y="VariableName", orientation='h')
        return feature_impact

    def summary_plot(self, df):
        df2 = self.data.summary_plot(df)

        summary_plot = px.scatter(df2, x="Feature Impact on Outcome", y="Feature Name", color="Rescaled Feature Value",
                                  hover_data=["Original Feature Value"], color_continuous_scale="RdBu")

        # return summary_plot, df2
        return df2

    def summary_plot_graph(self, df):
        summary_plot = px.scatter(df, x="Feature Impact on Outcome", y="Feature Name", color="Rescaled Feature Value",
                                  hover_data=["Original Feature Value"], color_continuous_scale="RdBu")
        return summary_plot

    def partial_dependence_plot(self, df, v1=None, v2=None, v3=None):
        pdp = shap_pdp()
        df = pdp.find(df)
        # add if and else so we can colors to discrete variables
        # g = px.scatter(df, x=v1, y=v2, color=v3, color_continuous_scale="RdBu",
        #                color_discrete_sequence=px.colors.sequential.Plasma_r)
        # # return g, df
        return df

    def pdp_plot(self, df, v1, v2, v3):
        g = px.scatter(df, x=v1, y=v2, color=v3, color_continuous_scale="RdBu",
                       color_discrete_sequence=px.colors.sequential.Plasma_r)
        return g

    def distributions(self, df, variable_name):
        graph = px.histogram(df, x=variable_name, marginal="box")
        return graph



from imports import *
from insight_classification import *
from insight_regression import *

"""
This class calculates feature importance

Input: 


"""


class insights():
    def __init__(self,param):
        super(insights, self).__init__()
        self.param = param
        self.regression = insight_regression()
        self.classification = insight_classification()

    def insight_1_feature_imp(self, df):
        if self.param["is_classification"] == True:
            return self.classification.insight_1_feature_imp(df)
        else:
            return self.regression.insight_1_feature_imp(df)

    def insight_2_global_feature_impact(self, df, outcome=0):
        if self.param["is_classification"] == True:
            return self.classification.insight_2_global_feature_impact(df, outcome, self.param["expected_values"],
                                                                       self.param["classes"])
        else:
            return self.regression.insight_2_global_feature_impact(df, self.param["expected_values"][0])
    
    def insight_2_local_feature_impact(self, df, outcome=0):
        if self.param["is_classification"] == True:
            return self.classification.insight_2_local_feature_impact(df, outcome, self.param["expected_values"],
                                                                       self.param["classes"])
        else:
            return self.regression.insight_2_global_feature_impact(df, self.param["expected_values"][0])

    def insight_3(self, df):
        if self.param["is_classification"] == True:
            return self.classification.insight_3(df)
        else:
            return self.regression.insight_3(df)






from imports import *
from plotly_graphs import *
from plotly_css import *
from app import app

def layout_interaction(original_variables, y_variables):
    layout =  html.Div([
                    dcc.Loading(
                        id="feature_interaction_load",
                        type="circle",
                        children=html.Div([
                            html.Div([
                                html.H4('Partial Dependence Plot',
                                        style=style12),
                                html.P(
                                    'The partial dependence plot (short PDP or PD plot) shows the marginal effect one or two features have on the predicted outcome of a machine learning model',
                                    style=style13),
                                html.Div([
                                    html.P('Variable Name'),
                                    dcc.Dropdown(
                                        id='xaxis-column',
                                        options=[{'label': i, 'value': i} for i in original_variables],
                                        value=original_variables[1],
                                        clearable=False
                                    ),

                                ],
                                    style={'width': '20%', 'marginLeft': 70, 'float': 'left',
                                            'display': 'inline-block'}),

                                html.Div([
                                    html.P("Variable Impact Values"),
                                    dcc.Dropdown(
                                        id='yaxis-column',
                                        options=[{'label': i, 'value': i} for i in y_variables],
                                        value=y_variables[1],
                                        clearable=False
                                    ),

                                ], style=style14),

                                html.Div([
                                    html.P('3rd Variable'),
                                    dcc.Dropdown(
                                        id='third-axis',
                                        options=[{'label': i, 'value': i} for i in original_variables],
                                        value=original_variables[-3],
                                        clearable=False
                                    ),

                                ], style=style15),

                            ]),
                            dcc.Loading(
                                id="loading-5",
                                type="circle",
                                children=dcc.Graph(id='indicator-graphic', style={'marginLeft': 50})
                            ),
                        ],
                            style=style16),
                    ),
                    dcc.Loading(
                        id="loading-2-pdp",
                        type='circle',
                        children=html.Div([
                            
                           
                            html.Div([
                                html.H4('Summary Plot',
                                        style=style17),
                                html.Div([
                                 dcc.Dropdown(
                                        id='xaxis-column-test',
                                        options=[{'label': i, 'value': i} for i in original_variables[1]],
                                        value=original_variables[1],
                                        clearable=False
                                    )
                            ], style={'display':'none'}),
                                html.P(
                                    'In the summary plot, we see first indications of the relationship between the value of a feature and the impact on the prediction',
                                    style=style18)
                                ,
                                # dcc.Loading(
                                #     id="loading-6",
                                #     type="circle",
                                #     children=
                                    dcc.Graph(id='summary_plot', style={'marginLeft': 50, 'height': '600px'})

                                # ),

                            ], style=style19),
                        ], style=style20)
                    )
                ]),
    return layout



from __future__ import print_function

from die import DIExplainer

from PDASH_utils import HeuristicSetSelection


class ProtodashExplainer(DIExplainer):
    """
    ProtodashExplainer provides exemplar-based explanations for summarizing datasets as well
    as explaining predictions made by an AI model. It employs a fast gradient based algorithm
    to find prototypes along with their (non-negative) importance weights. The algorithm minimizes the maximum
    mean discrepancy metric and has constant factor approximation guarantees for this weakly submodular function. [#]_.

    References:
        .. [#] `Karthik S. Gurumoorthy, Amit Dhurandhar, Guillermo Cecchi,
           "ProtoDash: Fast Interpretable Prototype Selection"
           <https://arxiv.org/abs/1707.01212>`_
    """

    def __init__(self):
        """
        Constructor method, initializes the explainer
        """
        super(ProtodashExplainer, self).__init__()

    def set_params(self, *argv, **kwargs):
        """
        Set parameters for the explainer.
        """
        pass

    def explain(self, X, Y, m, kernelType='other', sigma=2):
        """
        Return prototypes for data X, Y.

        Args:
            X (double 2d array): Dataset to select prototypical explanations from.
            Y (double 2d array): Dataset you want to explain.
            m (int): Number of prototypes
            kernelType (str): Type of kernel (viz. 'Gaussian', / 'other')
            sigma (double): width of kernel

        Returns:
            m selected prototypes from X and their (unnormalized) importance weights
        """
        return( HeuristicSetSelection(X, Y, m, kernelType, sigma) )



from imports import *


class draw_dataframe():
    def __init__(self):
        super(draw_dataframe, self).__init__()
        self.param= None


    def draw(self,  dataframe, max_rows=10):
        return self.generate_table(dataframe, max_rows)


    # Generating table from pandas dataframe
    def generate_table(self, dataframe, max_rows):
        return html.Table([
            html.Thead(
                html.Tr([html.Th(col) for col in dataframe.columns])
            ),
            html.Tbody([
                html.Tr([
                    html.Td(dataframe.iloc[i][col]) for col in dataframe.columns
                ]) for i in range(min(len(dataframe), max_rows))
            ])
        ])




from __future__ import print_function
import numpy as np
from sklearn.preprocessing import OneHotEncoder
from cvxopt.solvers import qp
from cvxopt import matrix, spmatrix
from numpy import array, ndarray
from scipy.spatial.distance import cdist


# solve_qp adapted from https://github.com/stephane-caron/qpsolvers/blob/master/qpsolvers/
# Added objective function value in return value

def cvxopt_matrix(M):
    if type(M) is ndarray:
        return matrix(M)
    elif type(M) is spmatrix or type(M) is matrix:
        return M
    coo = M.tocoo()
    return spmatrix(
        coo.data.tolist(), coo.row.tolist(), coo.col.tolist(), size=M.shape)


def solve_qp(P, q, G=None, h=None, A=None, b=None, solver=None, initvals=None):
    """
    Solve a Quadratic Program defined as:
        minimize
            (1/2) * x.T * P * x + q.T * x
        subject to
            G * x <= h
            A * x == b
    using CVXOPT <http://cvxopt.org/>.
    Parameters
    ----------
    P : numpy.array, cvxopt.matrix or cvxopt.spmatrix
        Symmetric quadratic-cost matrix.
    q : numpy.array, cvxopt.matrix or cvxopt.spmatrix
        Quadratic-cost vector.
    G : numpy.array, cvxopt.matrix or cvxopt.spmatrix
        Linear inequality matrix.
    h : numpy.array, cvxopt.matrix or cvxopt.spmatrix
        Linear inequality vector.
    A : numpy.array, cvxopt.matrix or cvxopt.spmatrix
        Linear equality constraint matrix.
    b : numpy.array, cvxopt.matrix or cvxopt.spmatrix
        Linear equality constraint vector.
    solver : string, optional
        Set to 'mosek' to run MOSEK rather than CVXOPT.
    initvals : numpy.array, optional
        Warm-start guess vector.
    Returns
    -------
    x : array, shape=(n,)
        Solution to the QP, if found, otherwise ``None``.
    Note
    ----
    CVXOPT only considers the lower entries of `P`, therefore it will use a
    wrong cost function if a non-symmetric matrix is provided.
    """
    args = [cvxopt_matrix(P), cvxopt_matrix(q)]
    if G is not None:
        args.extend([cvxopt_matrix(G), cvxopt_matrix(h)])
        if A is not None:
            if type(A) is ndarray and A.ndim == 1:
                A = A.reshape((1, A.shape[0]))
            args.extend([cvxopt_matrix(A), cvxopt_matrix(b)])
    sol = qp(*args, solver=solver, initvals=initvals)
    if 'optimal' not in sol['status']:
        return (None, None)
    return (  array(sol['x']).reshape((q.shape[0],)), sol['primal objective']  )



def runOptimiser(K, u, preOptw, initialValue, maxWeight=10000):
    """
    Args:
        K (double 2d array): Similarity/distance matrix
        u (double array): Mean similarity of each prototype
        preOptw (double): Weight vector
        initialValue (double): Initialize run
        maxWeight (double): Upper bound on weight

    Returns:
        Prototypes, weights and objective values
    """
    d = u.shape[0]
    lb = np.zeros((d, 1))
    ub = maxWeight * np.ones((d, 1))
    x0 = np.append( preOptw, initialValue/K[d-1, d-1] )

    G = np.vstack((np.identity(d), -1*np.identity(d)))
    h = np.vstack((ub, -1*lb))

    (sol, obj_value) = solve_qp(K, -u, G, h, A=None, b=None, solver=None, initvals=x0)
    return(sol, obj_value)




def get_Processed_NHANES_Data(filename):
    """
    Args:
        filename (str): Enter NHANES filename

    Returns:
        One hot encoded features and original input
    """
    # returns original and one hot encoded data
    # Input: XPT filename e.g. 2_H.XPT)
    # output:
    # One hot endcoded, e.g. (5924 x 145)
    # original, e.g. (5924 x 9)

    with open(filename, 'rb') as f:
        original = xport.to_numpy(f)

    # replace nan's with 0's.
    original[np.isnan(original)] = 0

    # delete 1st column (contains sequence numbers)
    original = original[:, 1:]

    # one hot encoding of all columns/features
    onehot_encoder = OneHotEncoder(sparse=False)
    onehot_encoded = onehot_encoder.fit_transform(original)

    # return one hot encoded and original data
    return (onehot_encoded, original)


def get_Gaussian_Data(nfeat, numX, numY):
    """
    Args:
        nfeat (int): Number of features
        numX (int): Size of X
        numY (int): Size of Y

    Returns:
        Datasets X and Y
    """
    np.random.seed(0)
    X = np.random.normal(0.0, 1.0, (numX, nfeat))
    Y = np.random.normal(0.0, 1.0, (numY, nfeat))

    for i in range(numX):
        X[i, :] = X[i, :] / np.linalg.norm(X[i, :])

    for i in range(numY):
        Y[i, :] = Y[i, :] / np.linalg.norm(Y[i, :])

    return(X, Y)


# expects X & Y in (observations x features) format

def HeuristicSetSelection(X, Y, m, kernelType, sigma):
    """
    Main prototype selection function.

    Args:
        X (double 2d array): Dataset to select prototypes from
        Y (double 2d array): Dataset to explain
        m (double): Number of prototypes
        kernelType (str): Gaussian, linear or other
        sigma (double): Gaussian kernel width

    Returns:
        Current optimum, the prototypes and objective values throughout selection
    """
    numY = Y.shape[0]
    numX = X.shape[0]
    allY = np.array(range(numY))

    # Store the mean inner products with X
    if kernelType == 'Gaussian':
        meanInnerProductX = np.zeros((numY, 1))
        for i in range(numY):
            Y1 = Y[i, :]
            Y1 = Y1.reshape(Y1.shape[0], 1).T
            distX = cdist(X, Y1)
            meanInnerProductX[i] = np.sum( np.exp(np.square(distX)/(-2.0 * sigma**2)) ) / numX
    else:
        M = np.dot(Y, np.transpose(X))
        meanInnerProductX = np.sum(M, axis=1) / M.shape[1]

    # move to features x observation format to be consistent with the earlier code version
    X = X.T
    Y = Y.T

    # Intialization
    S = np.zeros(m, dtype=int)
    setValues = np.zeros(m)
    sizeS = 0
    currSetValue = 0.0
    currOptw = np.array([])
    currK = np.array([])
    curru = np.array([])
    runningInnerProduct = np.zeros((m, numY))

    while sizeS < m:

        remainingElements = np.setdiff1d(allY, S[0:sizeS])

        newCurrSetValue = currSetValue
        maxGradient = 0

        for count in range(remainingElements.shape[0]):

            i = remainingElements[count]
            newZ = Y[:, i]

            if sizeS == 0:

                if kernelType == 'Gaussian':
                    K = 1
                else:
                    K = np.dot(newZ, newZ)

                u = meanInnerProductX[i]
                w = np.max(u / K, 0)
                incrementSetValue = -0.5 * K * (w ** 2) + (u * w)

                if (incrementSetValue > newCurrSetValue) or (count == 1):
                    # Bookeeping
                    newCurrSetValue = incrementSetValue
                    desiredElement = i
                    newCurrOptw = w
                    currK = K

            else:
                recentlyAdded = Y[:, S[sizeS - 1]]

                if kernelType == 'Gaussian':
                    distnewZ = np.linalg.norm(recentlyAdded-newZ)
                    runningInnerProduct[sizeS - 1, i] = np.exp( np.square(distnewZ)/(-2.0 * sigma**2 ) )
                else:
                    runningInnerProduct[sizeS - 1, i] = np.dot(recentlyAdded, newZ)

                innerProduct = runningInnerProduct[0:sizeS, i]
                if innerProduct.shape[0] > 1:
                    innerProduct = innerProduct.reshape((innerProduct.shape[0], 1))

                gradientVal = meanInnerProductX[i] - np.dot(currOptw, innerProduct)

                if (gradientVal > maxGradient) or (count == 1):
                    maxGradient = gradientVal
                    desiredElement = i
                    newinnerProduct = innerProduct[:]

        S[sizeS] = desiredElement

        curru = np.append(curru, meanInnerProductX[desiredElement])

        if sizeS > 0:

            if kernelType == 'Gaussian':
                selfNorm = array([1.0])
            else:
                addedZ = Y[:, desiredElement]
                selfNorm = array( [np.dot(addedZ, addedZ)] )

            K1 = np.hstack((currK, newinnerProduct))

            if newinnerProduct.shape[0] > 1:
                selfNorm = selfNorm.reshape((1,1))
            K2 = np.vstack( (K1, np.hstack((newinnerProduct.T, selfNorm))) )

            currK = K2
            if maxGradient <= 0:
                #newCurrOptw = np.vstack((currOptw[:], np.array([0])))
                newCurrOptw = np.append(currOptw, [0], axis=0)
                newCurrSetValue = currSetValue
            else:
                [newCurrOptw, value] = runOptimiser(currK, curru, currOptw, maxGradient)
                newCurrSetValue = -value

        currOptw = newCurrOptw
        if type(currOptw) != np.ndarray:
            currOptw = np.array([currOptw])

        currSetValue = newCurrSetValue

        setValues[sizeS] = currSetValue
        sizeS = sizeS + 1

    return(currOptw, S, setValues)








from imports import *
from encode_decode_cat_col import *
from rescale_numeric_feature import *
from PDASH import ProtodashExplainer
from PDASH_utils import *
"""
This class calculates similar prototypes

Input: 


"""


class protodash():
    def __init__(self):
        super(protodash, self).__init__()
        self.encode= None
        self.get_col=None
        self.cat_col=[]
        self.actual_variables= []
        self.classification= False
        self.df=None
        self.y_variable= None



    def encode_categorical_var(self, df):
        if self.get_col==None:
            self.get_col= get_cols()
            num_col, self.cat_col = self.get_col.get_cate_numer_col(df)

        if self.encode==None:
            self.encode= encode_decode_cat_col()

        for col_name in self.cat_col:
            df = self.encode.encode_col(df, col_name)

        return df

    def decode_categorical_var(self, df):
        # let's decode
        for col_name in self.cat_col:
            df = self.encode.decode_col(df, col_name)

        return df



    def preprocess_data(self,  df, y_variable):

        #store y_var
        self.y_variable=y_variable

        # encode categorical var
        df= self.encode_categorical_var(df)
        
        # check and remove index column in the file.
        try:
            df = df.drop("index", axis=1)
        except:
            pass

        # finad and remove variables containing shap values
        self.df= self.remove_impact_columns(df)

        self.classification= self.is_classification()

        return True




    def remove_impact_columns(self, df):
        self.actual_variables = [col for col in df.columns if not '_impact' in col]
        self.actual_variables = [col for col in self.actual_variables if not '_rescaled' in col]

        df = df[self.actual_variables]
        return df


    def is_classification(self):
        count = self.df[self.y_variable].nunique()

        if count>=10:
            return False
        else:
            return True

    def find_prototypes(self, row_number):

        # get training data
        data, Z= self.z_train_good(row_number)
        #get prototypes
        explainer = ProtodashExplainer()
        try:
            (W, S, setValues) = explainer.explain(Z, data, m=5, kernelType='other', sigma=5)
        except:
            (W, S, setValues) = explainer.explain(Z, data, m=5, kernelType='other', sigma=5) #Guassian gives an error. 
        #make a dataframe
        dfs = pd.DataFrame.from_records(data[S, 0:].astype('double'))
        dfs.columns = self.actual_variables
        dfs["Weight(%)"] = ((np.around(W, 5) / np.sum(np.around(W, 5)))*100)
        dfs = self.decode_categorical_var(dfs)
        return dfs,self.df.iloc[row_number]

    def z_train_good(self, row_number):
        row= self.df.iloc[row_number]
        #remove row from the data
        df = self.df
        df = self.df.drop(self.df.index[row_number])
        if self.classification==True:
            predict_value= row[self.y_variable]
            df= df[df[self.y_variable] == predict_value]  # choose prediction: 0/1
        row1 = row.values
        array= df.values
        Z= row1.reshape(-1, 1).T

        return array, Z






import os
import sys
from pathlib import Path
from sys import platform
import subprocess

path= Path(__file__).parent.absolute()
path_dataset= os.path.join(path, "datasets")
path= os.path.join(path, "lib")

sys.path.append(path)


from imports import *
from dashboard import *
from calculate_shap import *
"""
This class calculates feature importance

Input: 


"""


class explain():
    def __init__(self):
        super(explain, self).__init__()
        self.param= {}

    # is classification function?

    def is_classification_given_y_array(self, y_test):
        is_classification = False
        total = len(y_test)
        total_unique = len(set(y_test))
        if total < 30:
            if total_unique < 10:
                is_classification = True
        else:
            if total_unique < 20:
                is_classification = True
        return is_classification


    def ai(self,  df,  y, model, model_name="xgboost", mode=None):
        y_variable= "y_actual"
        y_variable_predict= "y_prediction"



        # If yes, then different shap functuions are required.
        # get the shap value based on predcton and make a new dataframe.

        # find predictions first as shap values need that.

        prediction_col=[]

        if model_name == "xgboost":
            import xgboost
            if xgboost.__version__ in ['1.1.0', '1.1.1', '1.1.0rc2', '1.1.0rc1']:
                print("Current Xgboost version is not supported. Please install Xgboost using 'pip install xgboost==1.0.2'")
                return False
            prediction_col = model.predict(xgboost.DMatrix(df))

        elif model_name == "catboost":
            prediction_col = model.predict(df.to_numpy())

        else:
            prediction_col = model.predict(df.to_numpy())

        # is classification?
        is_classification = self.is_classification_given_y_array(prediction_col)



        #shap
        c = calculate_shap()
        self.df_final, self.explainer = c.find(model, df, prediction_col, is_classification, model_name=model_name)

        #prediction col
        self.df_final[y_variable_predict] = prediction_col



        self.df_final[y_variable] = y


        #additional inputs.
        if is_classification==True:
            # find and add probabilities in the dataset.
            prediction_col_prob = model.predict_proba(df.to_numpy())
            pd_prediction_col_prob = pd.DataFrame(prediction_col_prob)

            for c in pd_prediction_col_prob.columns:
                self.df_final["probability_of_predicting_class_" + str(c)] = list(pd_prediction_col_prob[c])

            classes = []
            for c in pd_prediction_col_prob.columns:
                classes.append(str(c))
            self.param["classes"]=classes

            try:
                expected_values_by_class = self.explainer.expected_value
            except:
                expected_values_by_class=[]
                for c in range(len(classes)):
                    expected_values_by_class.append(1/len(classes))


            self.param["expected_values"]= expected_values_by_class
        else:
            try:
                expected_values = self.explainer.expected_value
                self.param["expected_values"] = [expected_values]
            except:
                expected_value = [round(np.array(y).mean(),2)]
                self.param["expected_values"] = expected_value


        self.param["is_classification"]= is_classification
        self.param["model_name"]= model_name
        self.param["model"]= model
        self.param["columns"]= df.columns
        self.param["y_variable"]= y_variable
        self.param["y_variable_predict"]= y_variable_predict


        d= dashboard()
        d.find(self.df_final, mode, self.param)

        return True

    def dataset_boston(self):
        # load JS visualization code to notebook
        shap.initjs()
        X, y = shap.datasets.boston()
        return X,y


    def dataset_iris(self):
        # load JS visualization code to notebook
        shap.initjs()
        X, y = shap.datasets.iris()
        return X,y


    def dataset_heloc(self):
        dataset= pd.read_csv(path_dataset+"/heloc_dataset.csv")

        map_riskperformance= {"RiskPerformance": {"Good":1, "Bad":0}}
        dataset.replace(map_riskperformance, inplace=True)
        y= list(dataset["RiskPerformance"])
        X= dataset.drop("RiskPerformance", axis=1)
        return X,y


    def run_only_first_time(self):

        if platform == "linux" or platform == "linux2":
            try:
                run_command("curl -sL https://rpm.nodesource.com/setup_10.x | sudo bash -")
                run_command("sudo apt install nodejs")
                run_command("sudo apt install npm")
            except:
                run_command("sudo yum install nodejs")
                run_command("sudo yum install npm")
            run_command("npm install -g localtunnel")

        elif platform == "darwin":
            run_command("xcode-select --install")
            run_command("brew install nodejs")
            run_command("npm install -g localtunnel")

        elif platform == "win32":
            print("Please install nodejs, npm, and localtunnel manually")
            run_command("npm install -g localtunnel")
        elif platform == "win64":
            print("Please install nodejs, npm, and localtunnel manually")
            run_command("npm install -g localtunnel")


explainx=explain()


def run_command(command):
    # subdomain= 'explainx-'+ get_random_string(10)
    command_arr = command.split(" ")

    task = subprocess.Popen(command_arr,
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE, preexec_fn=os.setsid)

    for line in iter(task.stdout.readline, b''):
        print('{0}'.format(line.decode('utf-8')), end='')






from imports import *
from rescale_numeric_feature import *
"""
This class helps to plot summary plot

Input: 


"""


class summary_plot():
    def __init__(self):
        super(summary_plot, self).__init__()
        self.param= None
        self.original_columns= None


    def find(self,  df):

        column = get_cols()
        self.original_columns = column.get_all_cols(df)

        re= rescale_numeric_features()
        df_with_rescaled_features= re.rescale(df)

        final_dataframe= self.rearrange_dataframe(df_with_rescaled_features )

        return final_dataframe

    def rearrange_dataframe(self, df_re ):

        df_final = pd.DataFrame()

        for v in self.original_columns:
            df_single = df_re[[v, v + '_rescaled', v + '_impact']]
            df_single["Variable Name"] = v
            df_single.columns = ['Original Feature Value', 'Rescaled Feature Value', 'Feature Impact on Outcome', 'Feature Name']
            df_final = pd.concat([df_final, df_single])

        return df_final








from imports import *

"""
This class calculates feature importance

Input: 


"""


class insight_regression():
    def __init__(self):
        super(insight_regression, self).__init__()
        self.param = None

    def insight_1_feature_imp(self, df):
        #
        top_important_variables = []
        average_shap_values = []

        df_numpy = df.to_numpy()

        for i in range(3):
            top_important_variables.append(df_numpy[-1 * i][0])
            average_shap_values.append(df_numpy[-1 * i][1])

        sentences = []
        sentences.append(
            "This graph helps you identify which features in your dataset have the greatest effect on the outcomes of your machine learning model.")
        sentences.append("In this graph, you can see: " + str(top_important_variables[0]) + " " +
                         str(top_important_variables[1]) + " " +
                         str(top_important_variables[
                                 2]) + " " + "are the top three most important variables according to your machine learning model.")

        sentences.append(
            "Remember, each variable might affect the outcome differently. Let’s scroll down and explore how each variable impacts the outcome.")

        sentences.append("On average, the variable " + str(top_important_variables[0]) + " " +
                         "will change the model outcome by " +
                         str(round(average_shap_values[0], 2)))

        return sentences

    def insight_2_global_feature_impact(self, df, expected_values):

        top_positive_variables = []
        average_shap_values_positive = []

        top_negative_variables = []
        average_shap_values_negative = []

        df_numpy = df.to_numpy()

        for i in range(3):
            top_positive_variables.append(df_numpy[-1 * i][0])
            average_shap_values_positive.append(df_numpy[-1 * i][1])

            top_negative_variables.append(df_numpy[i][0])
            average_shap_values_negative.append(df_numpy[i][1])

        sentences = []

        sentences.append("The average outcome predicted by the model is " + str(expected_values))

        sentences.append("On average, the variable " +
                         top_positive_variables[0] + ", " +
                         top_positive_variables[1] + ",and " +
                         top_positive_variables[2] + " " +
                         "will increase the average model outcome(" + str(expected_values) + ")" +
                         " by " +
                         str(round(average_shap_values_positive[0], 2)) + ", " +
                         str(round(average_shap_values_positive[1], 2)) + ", and " +
                         str(round(average_shap_values_positive[2], 2)) + " respectively.")

        sentences.append("On average, the variable " +
                         top_negative_variables[0] + ", " +
                         top_negative_variables[1] + ",and " +
                         top_negative_variables[2] + " " +
                         "will change the average model outcome(" + str(expected_values) + ")" +
                         " by " +
                         str(round(average_shap_values_negative[0], 2)) + ", " +
                         str(round(average_shap_values_negative[1], 2)) + ", and" +
                         str(round(average_shap_values_negative[2], 2)) + " respectively.")

        sentences.append("")

        return sentences

    def insight_3(self, df):

        top_positive_variables = []
        average_shap_values_positive = []

        df_numpy = df.to_numpy()

        for i in range(3):
            top_positive_variables.append(df_numpy[-1 * i][0])
            average_shap_values_positive.append(df_numpy[-1 * i][1])

        sentences = []

        sentences.append("Please explore the PDP graph to identify how different values of " +
                         top_positive_variables[0] + ", " +
                         top_positive_variables[1] + ", and " +
                         top_positive_variables[2] + " " +
                         " affect the model decision.")

        return sentences

    def insight_4_pdp(self, df):
        output = True
        return output

    def insight_5(self, x):
        output = True
        return output









from jupyter_dash import JupyterDash
import dash_bootstrap_components as dbc
import dash
import dash_core_components as dcc
import dash_core_components
import dash_html_components as html
from dash.dependencies import Input, Output, State
import plotly.express as px
import plotly.graph_objects as go
from plotly.graph_objs.scatter.marker import Line
import plotly.figure_factory as ff
import dash_table
import pandas as pd
import numpy as np
import dash_bootstrap_components as dbc
import time
import shap
import dash_editor_components
import socket
from contextlib import closing
import subprocess
import os
import sys
import shap

import random
import string



from sys import platform
import os, sys
from setuptools import setup, Extension


with open("README.md", "r") as fh:
    long_description = fh.read()




setup(
  name = 'explainx',         # How you named your package folder (MyLib)
  packages = ['explainx'],   # Chose the same as "name"
  version = '2.387',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Explainable AI Framework for Model Developers.',   # Give a short description about your library
  long_description=long_description,
  long_description_content_type="text/markdown",
  author = 'explainx.ai',                   # Type in your name
  author_email = 'muddassar@explainx.ai',      # Type in your E-Mail
  url = 'https://github.com/explainX/explainx',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/explainX/explainx/archive/v2.36.zip',    # I explain this later on
  keywords = ['Explainable AI', 'Explainable Machine Learning', 'trust', "interpretability", "transparent"],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'jupyter_dash==0.2.1.post1',
          'dash_bootstrap_components==0.10.2',
          'dash==1.12.0',
          'dash_core_components==1.10.0',
          'dash_html_components==1.0.3',
          'plotly==4.5.1',
          'dash_table==4.7.0',
          'pandas==1.0.4',
          'numpy==1.18.1',
          'dash_bootstrap_components==0.10.2',
          'cvxopt==1.2.4',
           'scikit-learn==0.22.1',
           'scipy==1.4.1',
            'pandasql==0.7.3',
            'tqdm==4.47.0',
            'dash-editor-components==0.0.2',
            'shap==0.34.0',
            'dash_daq==0.5.0'

      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
    package_data={"explainx":["lib/*", "lib/apps/*","tutorials/*", "datasets/*", "lib/shap/*","lib/shap/plots/*", "lib/shap/plots/resources/*",
                              "lib/shap/explainers/*",
                              "lib/shap/explainers/deep/*",
                              "lib/shap/explainers/other/*",
                              "lib/shap/benchmark/*" ]},
)



from imports import *
from plotly_graphs import *
from plotly_css import *
from app import app

def global_explanation(original_variables):
    layout =  html.Div(children=[
        

                            html.Div([
                                dcc.Loading([
                                    html.Div([
                                    dcc.Dropdown(
                                            id='xaxis-column-test-2',
                                            options=[{'label': i, 'value': i} for i in original_variables[1]],
                                            value=original_variables[1],
                                            clearable=False
                                        )
                                ], style={'display':'none'})]),
                                html.Div([
                                    html.H4('Global Feature Importance',
                                            style=style5),
                                    html.P(
                                        'Feature importance assign a score to input features based on how useful they are at predicting a target variable. ',
                                        style=style6)
                                    ,
                                    dcc.Loading(
                                        id="loading-1",
                                        type="circle",
                                        children=dbc.Row(
                                            [
                                                dbc.Col(html.Div(dcc.Graph(id="global_feature_importance",
                                                                        style={'marginLeft': 50, 'marginTop': 0,
                                                                                'height': '500px'}
                                                                        )), width=9),
                                                dbc.Col(
                                                    [
                                                        html.Div([
                                                            html.H2("How to read this graph?"),
                                                            html.P(
                                                                "This graph helps you identify which features in your dataset have the greatest effect on the outcomes of your machine learning model")
                                                        ]),
                                                        html.Div([
                                                            html.H2("Insights"),
                                                            html.P(id='global_message_1'),
                                                            html.P(id='global_message_2'),
                                                            html.P(id='global_message_3'),
                                                            html.P(id='global_message_4'),
                                                            html.P(id='global_message_5')

                                                        ]

                                                        )]
                                                )

                                            ])

                                    )
                                    , ], style=style7,
                                ),

                                html.Div([
                                    html.H4('Global Feature Impact',
                                            style=style8),
                                    html.P(
                                        'Feature impact identifies which features (also known as columns or inputs) in a dataset have the greatest positive or negative effect on the outcomes of a machine learning model.',
                                        style=style9),
                                    dcc.Loading(
                                        id="loading-2",
                                        type="circle",
                                        children=dbc.Row(
                                            [
                                                dbc.Col(html.Div(dcc.Graph(id='global_feature_impact',
                                                                        style={'marginLeft': 50, 'marginTop': 0,
                                                                                'height': '500px'})), width=9),
                                                dbc.Col(
                                                    [
                                                        html.Div([
                                                            html.H2("How to read this graph?"),
                                                            html.P(
                                                                "This tells you which features have positive impact and which features have negative impact on the output of the decision")
                                                        ]),
                                                        html.Div([
                                                            html.H2("Insights"),
                                                            html.P(id='message_1'),
                                                            html.P(id='message_2'),
                                                            html.P(id='message_3')
                                                        ]

                                                        )]
                                                )

                                            ])

                                    )

                                ],
                                    style=style10,
                                ),
                            ],

                                style=style11
                            )
                        ], style={'height': '400'})
    return layout



# Import modules
import pandas as pd
import numpy as np
import sklearn
from sklearn import *
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import label_binarize
from sklearn.metrics import * 
from itertools import cycle
import colorlover as cl
import plotly.figure_factory as ff
import plotly.graph_objects as go

class performance_metrics():
    def __init__(self, y_test, y_pred):
        super(performance_metrics, self).__init__()
        self.param = None
        self.y_test = y_test
        self.y_pred = y_pred


    def false_possitives_negatives(self):
        
        f1 = f1_score(self.y_test, self.y_pred, average='micro')
        accuracy = accuracy_score(self.y_test, self.y_pred)
        cnf_matrix = confusion_matrix(self.y_test, self.y_pred)

        FP = cnf_matrix.sum(axis=0) - np.diag(cnf_matrix) 
        FN = cnf_matrix.sum(axis=1) - np.diag(cnf_matrix)
        TP = np.diag(cnf_matrix)
        TN = cnf_matrix.sum() - (FP + FN + TP)
        FP = FP.astype(float).sum()
        FN = FN.astype(float).sum()
        TP = TP.astype(float).sum()
        TN = TN.astype(float).sum()

        # Sensitivity, hit rate, recall, or true positive rate
        TPR = TP/(TP+FN)
        # Specificity or true negative rate
        TNR = TN/(TN+FP) 
        # Precision or positive predictive value
        PPV = TP/(TP+FP)
        # Negative predictive value
        NPV = TN/(TN+FN)
        # Fall out or false positive rate
        FPR = FP/(FP+TN)
        # False negative rate
        FNR = FN/(TP+FN)
        # False discovery rate
        FDR = FP/(TP+FP)
        # Overall accuracy for each class
        ACC = (TP+TN)/(TP+FP+FN+TN)


        fp_fn_table = pd.DataFrame(dict({'False Possitives (%)': [(FPR * 100).round(2)],
                                         'False Negatives (%)': [(FNR * 100).round(2)],
                                         'Accuracy (%)': accuracy.round(2)* 100,
                                         'F1': f1.round(2)}))
        
        fig_metrics_table = ff.create_table(fp_fn_table, height_constant=15)
        
        return fig_metrics_table    
    
        
        
    
    def get_matrix(self):
        y_test = label_binarize(self.y_test, classes=list(set(self.y_test.flatten())))
        n_classes = y_test.shape[1]
        conf_matrix = confusion_matrix(y_test.argmax(axis=1), self.y_pred.argmax(axis=1))
        
        return conf_matrix 
    
    
    def plot_roc(self):
    
        y_test = label_binarize(self.y_test, classes=list(set(self.y_test.flatten())))
        n_classes = y_test.shape[1]

        # Compute ROC curve and ROC area for each class
        fpr = dict()
        tpr = dict()
        roc_auc = dict()
        for i in range(n_classes):
            fpr[i], tpr[i], _ = roc_curve(y_test[:, i], self.y_pred[:, i])
            roc_auc[i] = auc(fpr[i], tpr[i])

        # Compute micro-average ROC curve and ROC area
        fpr["micro"], tpr["micro"], _ = roc_curve(y_test.ravel(), self.y_pred.ravel())
        roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])

        # Compute macro-average ROC curve and ROC area

        # First aggregate all false positive rates
        all_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))

        # Then interpolate all ROC curves at this points
        mean_tpr = np.zeros_like(all_fpr)
        for i in range(n_classes):
            mean_tpr += np.interp(all_fpr, fpr[i], tpr[i])

        # Finally average it and compute AUC
        mean_tpr /= n_classes

        fpr["macro"] = all_fpr
        tpr["macro"] = mean_tpr
        roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])

        # Plot all ROC curves
        data = []
        trace1 = go.Scatter(x=fpr["micro"], y=tpr["micro"], mode='lines', 
                            line=dict(color='deeppink', width=2, dash='dot'),
                            name='micro-average ROC curve (area = {0:0.2f})'
                                   ''.format(roc_auc["micro"]))
        data.append(trace1)

        trace2 = go.Scatter(x=fpr["macro"], y=tpr["macro"], mode='lines', 
                            line=dict(color='navy', width=2, dash='dot'),
                            name='macro-average ROC curve (area = {0:0.2f})'
                                  ''.format(roc_auc["macro"]))
        data.append(trace2)

        colors = cycle(['#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A', '#19D3F3', '#FF6692', '#B6E880', '#FF97FF', '#FECB52'])

        for i, color in zip(range(n_classes), colors):
            trace3 = go.Scatter(x=fpr[i], y=tpr[i],
                                mode='lines', 
                                line=dict(color=color, width=2),
                                name='ROC curve of class {0} (area = {1:0.2f})'
                                ''.format(i, roc_auc[i]))
            data.append(trace3)

        trace4 = go.Scatter(x=[0, 1], y=[0, 1], 
                            mode='lines', 
                            line=dict(color='black', width=2, dash='dash'),
                            showlegend=False)


        layout = go.Layout(title='Receiver operating characteristic',
                           xaxis=dict(title='False Positive Rate'),
                           yaxis=dict(title='True Positive Rate'),
                           margin=dict(pad=25))

        fig = go.Figure(data=data, layout=layout)

        return fig
    
    
    def plot_pr(self):
        
        colors = cycle(['#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A',
                        '#19D3F3', '#FF6692', '#B6E880', '#FF97FF', '#FECB52'])

        # Binarize the output
        y_test = label_binarize(self.y_test, classes=list(set(self.y_test.flatten())))
        n_classes = y_test.shape[1]

        # Compute Precision-Recall and plot curve
        precision = dict()
        recall = dict()
        average_precision = dict()
        
        for i in range(n_classes):
            precision[i], recall[i], _ = precision_recall_curve(y_test[:, i], self.y_pred[:, i])
            average_precision[i] = average_precision_score(y_test[:, i], self.y_pred[:, i])

        # Compute micro-average ROC curve and ROC area
        precision["micro"], recall["micro"], _ = precision_recall_curve(y_test.ravel(),
            self.y_pred.ravel())
        average_precision["micro"] = average_precision_score(y_test, self.y_pred,
                                                             average="micro")


        data = []
        trace2 = go.Scatter(x=recall["micro"], y=precision["micro"], 
                            mode='lines',
                            line=dict(color='gold', width=2),
                            name='micro-average Precision-recall curve (area = {0:0.2f})'
                                  ''.format(average_precision["micro"]))
        data.append(trace2)
        
        for i, color in zip(range(n_classes), colors):
            trace3 = go.Scatter(x=recall[i], y=precision[i],
                                mode='lines',
                                line=dict(color=color, width=2),
                                name='Precision-recall curve of class {0} (area = {1:0.2f})'
                                      ''.format(i, average_precision[i]))
            data.append(trace3)

        layout = go.Layout(title='Precision-Recall curve',
                           xaxis=dict(title='Recall'),
                           yaxis=dict(title='Precision'),
                           margin=dict(pad=25))

        fig = go.Figure(data=data, layout=layout)
        
        return fig
    
    
    def plot_pie(self):
        
        cnf_matrix = confusion_matrix(self.y_test, self.y_pred)

        FP = cnf_matrix.sum(axis=0) - np.diag(cnf_matrix) 
        FN = cnf_matrix.sum(axis=1) - np.diag(cnf_matrix)
        TP = np.diag(cnf_matrix)
        TN = cnf_matrix.sum() - (FP + FN + TP)
        fp = FP.astype(float).sum()
        fn = FN.astype(float).sum()
        tp = TP.astype(float).sum()
        tn = TN.astype(float).sum()
        
        
        label_text = ["True Positive", "False Negative","False Positive","True Negative"]
        labels = ["TP", "FN", "FP", "TN"]
        blue = cl.flipper()['seq']['9']['Blues']
        red = cl.flipper()['seq']['9']['Reds']
        colors = [blue[4], blue[1], red[1], red[4]]

        trace0 = go.Pie(
        labels=label_text,
        values=[tp, fn, fp, tn],
        hoverinfo='label+value+percent',
        textinfo='text+value',
        text=labels,
        sort=False,
        marker=dict(
        colors=colors))

        layout = go.Layout(
        title=f'TP, TN, FP, FN',
        margin=dict(l=10, r=10, t=60, b=10),
        legend=dict(
        bgcolor='rgba(255,255,255,0)',
        orientation='h'))

        data = [trace0]
        figure = go.Figure(data=data, layout=layout)
        return figure
    
    
    def plot_matrix(self):
        
        matrix_data =  pd.DataFrame(confusion_matrix(self.y_test, self.y_pred))
        matrix_data = matrix_data.astype('float') / matrix_data.sum(axis=1)[:, np.newaxis]*100

        z = np.array(matrix_data).round(2)

        x = list(matrix_data.columns)
        y =  list(matrix_data.index)

        # change each element of z to type string for annotations
        z_text = [[str(y) for y in x] for x in z]

        # set up figure 
        fig_matrix = ff.create_annotated_heatmap(z, x=x, y=y, annotation_text=z_text, colorscale='Viridis')

        # add title
        fig_matrix.update_layout(title_text='<i><b>Confusion matrix (%)</b></i>',
                          xaxis = dict(title='Predicted Label'),
                          yaxis = dict(title='True Label'))

        # adjust margins to make room for yaxis title
        fig_matrix.update_layout(margin=dict(t=50, l=200))

        # add colorbar
        fig_matrix['data'][0]['showscale'] = True
        
        return fig_matrix
    
    
    def metrics_table(self):
       
        
                 # AUC score
        lb = preprocessing.LabelBinarizer()
        lb.fit(self.y_test)
        y_test_roc = lb.transform(self.y_test)
        y_pred_roc = lb.transform(self.y_pred)
        auc = roc_auc_score(y_test_roc, y_pred_roc, average="macro").round(2)

                    # MAE
        mae = mean_absolute_error(self.y_test, self.y_pred).round(2)
                # MSE
        mse = mean_squared_error(self.y_test, self.y_pred).round(2)

        # Accuracy
        accuracy = accuracy_score(self.y_test, self.y_pred).round(2)   

        # matthews_corrcoef
        matthews_corrcoe = matthews_corrcoef(self.y_test, self.y_pred).round(2)

    
       
        # Make DataFrames
        metric = ['Accuracy','Area Under Curve','MAE','MSE','Matthews Corrcoef']
        values = [accuracy, auc, mae, mse,matthews_corrcoe]
        metrics_tab = pd.DataFrame({'metric': metric, 'values': values})
        
        fig_metrics_table = ff.create_table(metrics_tab, height_constant=15)
        
   
        return fig_metrics_table



    def performance_metrics_regression(self):
        
        
        
        maxerror = max_error(self.y_test, self.y_pred).round(2) 

        mae = mean_absolute_error(self.y_test, self.y_pred).round(2) 
      
        mse = mean_squared_error(self.y_test, self.y_pred).round(2) 

        r2 = r2_score(self.y_test, self.y_pred).round(2) 
     
    
        def mean_absolute_percentage_error(y_true, y_pred): 
            y_true, y_pred = np.array(y_true), np.array(y_pred)
            return np.mean(np.abs((y_true - y_pred) / y_true)) * 100

        mape = mean_absolute_percentage_error(self.y_test, self.y_pred).round(2) 
    
        
        # Make DataFrames
        metric = ['Max Error','R squared','MAE','MSE','MAPE']
        values = [maxerror, r2, mae, mse, mape]
        metrics_dataframe = pd.DataFrame({'metric': metric, 'values': values})
        metrics_dataframe = ff.create_table(metrics_dataframe, height_constant=15)
        return metrics_dataframe

    def actuals(self):
        
        
        cnf_matrix = confusion_matrix(self.y_test, self.y_pred)

        FP = cnf_matrix.sum(axis=0) - np.diag(cnf_matrix) 
        FN = cnf_matrix.sum(axis=1) - np.diag(cnf_matrix)
        TP = np.diag(cnf_matrix)
        TN = cnf_matrix.sum() - (FP + FN + TP)
        fp = FP.astype(float).sum().round()
        fn = FN.astype(float).sum().round()
        tp = TP.astype(float).sum().round()
        tn = TN.astype(float).sum().round()

        total = tn + fn + fp + tp

        predicted_no = [tn,fn]
        predicted_yes = [fp,tp]


        data = {'Predicted No':predicted_no, 'Predicted Yes':predicted_yes} 

        # Creates pandas DataFrame. 
        df = pd.DataFrame(data, index =['No', 'Yes'])
        df['Total'] = df.sum(axis=1)
        df.loc['Total',:]= df.sum(axis=0)
        df.reset_index(inplace=True)
        df.rename({"index" : 'Actual'}, axis=1, inplace=True)

        actuals_table = ff.create_table(df, height_constant=15)

        return actuals_table